@@ -28,22 +28,11 @@ from random import Random
 
 __all__ = ("ResourcePath", "ResourcePathExpression")
 
-from typing import (
-    TYPE_CHECKING,
-    Any,
-    Dict,
-    Iterable,
-    Iterator,
-    List,
-    Literal,
-    Optional,
-    Tuple,
-    Type,
-    Union,
-    overload,
-)
+from collections.abc import Iterable, Iterator
+from typing import TYPE_CHECKING, Any, Literal, overload
 
 from ._resourceHandles._baseResourceHandle import ResourceHandleProtocol
+from .utils import ensure_directory_is_writeable
 
 if TYPE_CHECKING:
     from .utils import TransactionProtocol
@@ -63,12 +52,7 @@ ESCAPED_HASH = urllib.parse.quote("#")
 MAX_WORKERS = 10
 
 
-ResourcePathExpression = Union[str, urllib.parse.ParseResult, "ResourcePath", Path]
-"""Type-annotation alias for objects that can be coerced to ResourcePath.
-"""
-
-
-class ResourcePath:
+class ResourcePath:  # numpydoc ignore=PR02
     """Convenience wrapper around URI parsers.
 
     Provides access to URI components and can convert file
@@ -79,33 +63,41 @@ class ResourcePath:
 
     Parameters
     ----------
-    uri : `str`, `Path`, `urllib.parse.ParseResult`, or `ResourcePath`.
-        URI in string form.  Can be scheme-less if referring to a local
-        filesystem path.
+    uri : `str`, `pathlib.Path`, `urllib.parse.ParseResult`, or `ResourcePath`
+        URI in string form.  Can be scheme-less if referring to a relative
+        path or an absolute path on the local file system.
     root : `str` or `ResourcePath`, optional
         When fixing up a relative path in a ``file`` scheme or if scheme-less,
         use this as the root. Must be absolute.  If `None` the current
-        working directory will be used. Can be a file URI.
+        working directory will be used. Can be any supported URI scheme.
+        Not used if ``forceAbsolute`` is `False`.
     forceAbsolute : `bool`, optional
         If `True`, scheme-less relative URI will be converted to an absolute
         path using a ``file`` scheme. If `False` scheme-less URI will remain
-        scheme-less and will not be updated to ``file`` or absolute path.
-    forceDirectory: `bool`, optional
+        scheme-less and will not be updated to ``file`` or absolute path unless
+        it is already an absolute path, in which case it will be updated to
+        a ``file`` scheme.
+    forceDirectory : `bool`, optional
         If `True` forces the URI to end with a separator, otherwise given URI
         is interpreted as is.
     isTemporary : `bool`, optional
         If `True` indicates that this URI points to a temporary resource.
         The default is `False`, unless ``uri`` is already a `ResourcePath`
         instance and ``uri.isTemporary is True``.
+
+    Notes
+    -----
+    A non-standard URI of the form ``file:dir/file.txt`` is always converted
+    to an absolute ``file`` URI.
     """
 
-    _pathLib: Type[PurePath] = PurePosixPath
+    _pathLib: type[PurePath] = PurePosixPath
     """Path library to use for this scheme."""
 
     _pathModule = posixpath
     """Path module to use for this scheme."""
 
-    transferModes: Tuple[str, ...] = ("copy", "auto", "move")
+    transferModes: tuple[str, ...] = ("copy", "auto", "move")
     """Transfer modes supported by this implementation.
 
     Move is special in that it is generally a copy followed by an unlink.
@@ -141,15 +133,24 @@ class ResourcePath:
     def __new__(
         cls,
         uri: ResourcePathExpression,
-        root: Optional[Union[str, ResourcePath]] = None,
+        root: str | ResourcePath | None = None,
         forceAbsolute: bool = True,
         forceDirectory: bool = False,
-        isTemporary: Optional[bool] = None,
+        isTemporary: bool | None = None,
     ) -> ResourcePath:
         """Create and return new specialist ResourcePath subclass."""
         parsed: urllib.parse.ParseResult
         dirLike: bool = False
-        subclass: Optional[Type[ResourcePath]] = None
+        subclass: type[ResourcePath] | None = None
+
+        # Force root to be a ResourcePath -- this simplifies downstream
+        # code.
+        if root is None:
+            root_uri = None
+        elif isinstance(root, str):
+            root_uri = ResourcePath(root, forceDirectory=True, forceAbsolute=True)
+        else:
+            root_uri = root
 
         if isinstance(uri, os.PathLike):
             uri = str(uri)
@@ -166,14 +167,21 @@ class ResourcePath:
                 if ESCAPES_RE.search(uri):
                     log.warning("Possible double encoding of %s", uri)
                 else:
-                    uri = urllib.parse.quote(uri)
-                    # Special case hash since we must support fragments
-                    # even in schemeless URIs -- although try to only replace
-                    # them in file part and not directory part
-                    if ESCAPED_HASH in uri:
+                    # Fragments are generally not encoded so we must search
+                    # for the fragment boundary ourselves. This is making
+                    # an assumption that the filename does not include a "#"
+                    # and also that there is no "/" in the fragment itself.
+                    to_encode = uri
+                    fragment = ""
+                    if "#" in uri:
                         dirpos = uri.rfind("/")
-                        # Do replacement after this /
-                        uri = uri[: dirpos + 1] + uri[dirpos + 1 :].replace(ESCAPED_HASH, "#")
+                        trailing = uri[dirpos + 1 :]
+                        hashpos = trailing.rfind("#")
+                        if hashpos != -1:
+                            fragment = trailing[hashpos:]
+                            to_encode = uri[: dirpos + hashpos + 1]
+
+                    uri = urllib.parse.quote(to_encode) + fragment
 
             parsed = urllib.parse.urlparse(uri)
         elif isinstance(uri, urllib.parse.ParseResult):
@@ -224,9 +232,39 @@ class ResourcePath:
         if subclass is None:
             # Work out the subclass from the URI scheme
             if not parsed.scheme:
-                from .schemeless import SchemelessResourcePath
+                # Root may be specified as a ResourcePath that overrides
+                # the schemeless determination.
+                if (
+                    root_uri is not None
+                    and root_uri.scheme != "file"  # file scheme has different code path
+                    and not parsed.path.startswith("/")  # Not already absolute path
+                ):
+                    if not root_uri.dirLike:
+                        raise ValueError(
+                            f"Root URI ({root}) was not a directory so can not be joined with"
+                            f" path {parsed.path!r}"
+                        )
+                    # If root is temporary or this schemeless is temporary we
+                    # assume this URI is temporary.
+                    isTemporary = isTemporary or root_uri.isTemporary
+                    joined = root_uri.join(
+                        parsed.path, forceDirectory=forceDirectory, isTemporary=isTemporary
+                    )
 
-                subclass = SchemelessResourcePath
+                    # Rather than returning this new ResourcePath directly we
+                    # instead extract the path and the scheme and adjust the
+                    # URI we were given -- we need to do this to preserve
+                    # fragments since join() will drop them.
+                    parsed = parsed._replace(scheme=joined.scheme, path=joined.path, netloc=joined.netloc)
+                    subclass = type(joined)
+
+                    # Clear the root parameter to indicate that it has
+                    # been applied already.
+                    root_uri = None
+                else:
+                    from .schemeless import SchemelessResourcePath
+
+                    subclass = SchemelessResourcePath
             elif parsed.scheme == "file":
                 from .file import FileResourcePath
 
@@ -259,7 +297,7 @@ class ResourcePath:
                 )
 
             parsed, dirLike = subclass._fixupPathUri(
-                parsed, root=root, forceAbsolute=forceAbsolute, forceDirectory=forceDirectory
+                parsed, root=root_uri, forceAbsolute=forceAbsolute, forceDirectory=forceDirectory
             )
 
             # It is possible for the class to change from schemeless
@@ -366,11 +404,11 @@ class ResourcePath:
         Returns
         -------
         uri : `ResourcePath`
-            root URI.
+            Root URI.
         """
         return self.replace(path="", forceDirectory=True)
 
-    def split(self) -> Tuple[ResourcePath, str]:
+    def split(self) -> tuple[ResourcePath, str]:
         """Split URI into head and tail.
 
         Returns
@@ -379,13 +417,13 @@ class ResourcePath:
             Everything leading up to tail, expanded and normalized as per
             ResourcePath rules.
         tail : `str`
-            Last `self.path` component. Tail will be empty if path ends on a
+            Last path component. Tail will be empty if path ends on a
             separator. Tail will never contain separators. It will be
             unquoted.
 
         Notes
         -----
-        Equivalent to `os.path.split()` where head preserves the URI
+        Equivalent to `os.path.split` where head preserves the URI
         components.
         """
         head, tail = self._pathModule.split(self.path)
@@ -414,7 +452,7 @@ class ResourcePath:
         If URI ends on a slash returns an empty string. This is the second
         element returned by `split()`.
 
-        Equivalent of `os.path.basename()``.
+        Equivalent of `os.path.basename`.
         """
         return self.split()[1]
 
@@ -429,7 +467,7 @@ class ResourcePath:
 
         Notes
         -----
-        Equivalent of `os.path.dirname()`.
+        Equivalent of `os.path.dirname`.
         """
         return self.split()[0]
 
@@ -496,6 +534,7 @@ class ResourcePath:
         Returns
         -------
         updated : `ResourcePath`
+            Updated `ResourcePath` with new updated final component.
 
         Notes
         -----
@@ -511,7 +550,7 @@ class ResourcePath:
         updated.dirLike = False
         return updated
 
-    def updatedExtension(self, ext: Optional[str]) -> ResourcePath:
+    def updatedExtension(self, ext: str | None) -> ResourcePath:
         """Return a new `ResourcePath` with updated file extension.
 
         All file extensions are replaced.
@@ -582,7 +621,7 @@ class ResourcePath:
         return ext
 
     def join(
-        self, path: Union[str, ResourcePath], isTemporary: Optional[bool] = None, forceDirectory: bool = False
+        self, path: str | ResourcePath, isTemporary: bool | None = None, forceDirectory: bool = False
     ) -> ResourcePath:
         """Return new `ResourcePath` with additional path components.
 
@@ -593,7 +632,7 @@ class ResourcePath:
             to include a file at the end. Will be quoted depending on the
             associated URI scheme. If the path looks like a URI with a scheme
             referring to an absolute location, it will be returned
-            directly (matching the behavior of `os.path.join()`). It can
+            directly (matching the behavior of `os.path.join`). It can
             also be a `ResourcePath`.
         isTemporary : `bool`, optional
             Indicate that the resulting URI represents a temporary resource.
@@ -673,7 +712,7 @@ class ResourcePath:
             isTemporary=isTemporary,
         )
 
-    def relative_to(self, other: ResourcePath) -> Optional[str]:
+    def relative_to(self, other: ResourcePath) -> str | None:
         """Return the relative path from this URI to the other URI.
 
         Parameters
@@ -697,12 +736,18 @@ class ResourcePath:
             other = other.abspath()
 
         # Scheme-less self is handled elsewhere.
-        if self.scheme != other.scheme or self.netloc != other.netloc:
+        if self.scheme != other.scheme:
             return None
+        if self.netloc != other.netloc:
+            # Special case for localhost vs empty string.
+            # There can be many variants of localhost.
+            local_netlocs = {"", "localhost", "localhost.localdomain", "127.0.0.1"}
+            if not {self.netloc, other.netloc}.issubset(local_netlocs):
+                return None
 
         enclosed_path = self._pathLib(self.relativeToPathRoot)
         parent_path = other.relativeToPathRoot
-        subpath: Optional[str]
+        subpath: str | None
         try:
             subpath = str(enclosed_path.relative_to(parent_path))
         except ValueError:
@@ -722,9 +767,40 @@ class ResourcePath:
         raise NotImplementedError()
 
     @classmethod
-    def mexists(cls, uris: Iterable[ResourcePath]) -> Dict[ResourcePath, bool]:
+    def mexists(cls, uris: Iterable[ResourcePath]) -> dict[ResourcePath, bool]:
         """Check for existence of multiple URIs at once.
 
+        Parameters
+        ----------
+        uris : iterable of `ResourcePath`
+            The URIs to test.
+
+        Returns
+        -------
+        existence : `dict` of [`ResourcePath`, `bool`]
+            Mapping of original URI to boolean indicating existence.
+        """
+        # Group by scheme to allow a subclass to be able to use
+        # specialized implementations.
+        grouped: dict[type, list[ResourcePath]] = {}
+        for uri in uris:
+            uri_class = uri.__class__
+            if uri_class not in grouped:
+                grouped[uri_class] = []
+            grouped[uri_class].append(uri)
+
+        existence: dict[ResourcePath, bool] = {}
+        for uri_class in grouped:
+            existence.update(uri_class._mexists(grouped[uri_class]))
+
+        return existence
+
+    @classmethod
+    def _mexists(cls, uris: Iterable[ResourcePath]) -> dict[ResourcePath, bool]:
+        """Check for existence of multiple URIs at once.
+
+        Implementation helper method for `mexists`.
+
         Parameters
         ----------
         uris : iterable of `ResourcePath`
@@ -738,7 +814,7 @@ class ResourcePath:
         exists_executor = concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKERS)
         future_exists = {exists_executor.submit(uri.exists): uri for uri in uris}
 
-        results: Dict[ResourcePath, bool] = {}
+        results: dict[ResourcePath, bool] = {}
         for future in concurrent.futures.as_completed(future_exists):
             uri = future_exists[future]
             try:
@@ -775,7 +851,7 @@ class ResourcePath:
         """
         return self
 
-    def _as_local(self) -> Tuple[str, bool]:
+    def _as_local(self) -> tuple[str, bool]:
         """Return the location of the (possibly remote) resource as local file.
 
         This is a helper function for `as_local` context manager.
@@ -836,7 +912,7 @@ class ResourcePath:
     @classmethod
     @contextlib.contextmanager
     def temporary_uri(
-        cls, prefix: Optional[ResourcePath] = None, suffix: Optional[str] = None
+        cls, prefix: ResourcePath | None = None, suffix: str | None = None
     ) -> Iterator[ResourcePath]:
         """Create a temporary file-like URI.
 
@@ -857,7 +933,13 @@ class ResourcePath:
         """
         use_tempdir = False
         if prefix is None:
-            prefix = ResourcePath(tempfile.mkdtemp(), forceDirectory=True, isTemporary=True)
+            directory = tempfile.mkdtemp()
+            # If the user has set a umask that restricts the owner-write bit,
+            # the directory returned from mkdtemp may not initially be
+            # writeable by us
+            ensure_directory_is_writeable(directory)
+
+            prefix = ResourcePath(directory, forceDirectory=True, isTemporary=True)
             # Record that we need to delete this directory. Can not rely
             # on isTemporary flag since an external prefix may have that
             # set as well.
@@ -882,12 +964,10 @@ class ResourcePath:
             if use_tempdir:
                 shutil.rmtree(prefix.ospath, ignore_errors=True)
             else:
-                try:
+                with contextlib.suppress(FileNotFoundError):
                     # It's okay if this does not work because the user removed
                     # the file.
                     temporary_uri.remove()
-                except FileNotFoundError:
-                    pass
 
     def read(self, size: int = -1) -> bytes:
         """Open the resource and return the contents in bytes.
@@ -979,14 +1059,14 @@ class ResourcePath:
         # Implement here because the __new__ method confuses things
         return self
 
-    def __getnewargs__(self) -> Tuple:
+    def __getnewargs__(self) -> tuple:
         """Support pickling."""
         return (str(self),)
 
     @classmethod
     def _fixDirectorySep(
         cls, parsed: urllib.parse.ParseResult, forceDirectory: bool = False
-    ) -> Tuple[urllib.parse.ParseResult, bool]:
+    ) -> tuple[urllib.parse.ParseResult, bool]:
         """Ensure that a path separator is present on directory paths.
 
         Parameters
@@ -1027,17 +1107,17 @@ class ResourcePath:
     def _fixupPathUri(
         cls,
         parsed: urllib.parse.ParseResult,
-        root: Optional[Union[str, ResourcePath]] = None,
+        root: ResourcePath | None = None,
         forceAbsolute: bool = False,
         forceDirectory: bool = False,
-    ) -> Tuple[urllib.parse.ParseResult, bool]:
+    ) -> tuple[urllib.parse.ParseResult, bool]:
         """Correct any issues with the supplied URI.
 
         Parameters
         ----------
         parsed : `~urllib.parse.ParseResult`
             The result from parsing a URI using `urllib.parse`.
-        root : `str` or `ResourcePath`, ignored
+        root : `ResourcePath`, ignored
             Not used by the this implementation since all URIs are
             absolute except for those representing the local file system.
         forceAbsolute : `bool`, ignored.
@@ -1076,7 +1156,7 @@ class ResourcePath:
         src: ResourcePath,
         transfer: str,
         overwrite: bool = False,
-        transaction: Optional[TransactionProtocol] = None,
+        transaction: TransactionProtocol | None = None,
     ) -> None:
         """Transfer to this URI from another.
 
@@ -1114,8 +1194,8 @@ class ResourcePath:
         raise NotImplementedError(f"No transfer modes supported by URI scheme {self.scheme}")
 
     def walk(
-        self, file_filter: Optional[Union[str, re.Pattern]] = None
-    ) -> Iterator[Union[List, Tuple[ResourcePath, List[str], List[str]]]]:
+        self, file_filter: str | re.Pattern | None = None
+    ) -> Iterator[list | tuple[ResourcePath, list[str], list[str]]]:
         """Walk the directory tree returning matching files and directories.
 
         Parameters
@@ -1139,7 +1219,7 @@ class ResourcePath:
     def findFileResources(
         cls,
         candidates: Iterable[ResourcePathExpression],
-        file_filter: Optional[Union[str, re.Pattern]],
+        file_filter: str | re.Pattern | None,
         grouped: Literal[True],
     ) -> Iterator[Iterator[ResourcePath]]:
         ...
@@ -1159,7 +1239,7 @@ class ResourcePath:
     def findFileResources(
         cls,
         candidates: Iterable[ResourcePathExpression],
-        file_filter: Optional[Union[str, re.Pattern]] = None,
+        file_filter: str | re.Pattern | None = None,
         grouped: Literal[False] = False,
     ) -> Iterator[ResourcePath]:
         ...
@@ -1168,9 +1248,9 @@ class ResourcePath:
     def findFileResources(
         cls,
         candidates: Iterable[ResourcePathExpression],
-        file_filter: Optional[Union[str, re.Pattern]] = None,
+        file_filter: str | re.Pattern | None = None,
         grouped: bool = False,
-    ) -> Iterator[Union[ResourcePath, Iterator[ResourcePath]]]:
+    ) -> Iterator[ResourcePath | Iterator[ResourcePath]]:
         """Get all the files from a list of values.
 
         Parameters
@@ -1236,7 +1316,7 @@ class ResourcePath:
         self,
         mode: str = "r",
         *,
-        encoding: Optional[str] = None,
+        encoding: str | None = None,
         prefer_file_temporary: bool = False,
     ) -> Iterator[ResourceHandleProtocol]:
         """Return a context manager that wraps an object that behaves like an
@@ -1246,13 +1326,13 @@ class ResourcePath:
         ----------
         mode : `str`
             String indicating the mode in which to open the file.  Values are
-            the same as those accepted by `builtins.open`, though intrinsically
+            the same as those accepted by `open`, though intrinsically
             read-only URI types may only support read modes, and
             `io.IOBase.seekable` is not guaranteed to be `True` on the returned
             object.
         encoding : `str`, optional
             Unicode encoding for text IO; ignored for binary IO.  Defaults to
-            ``locale.getpreferredencoding(False)``, just as `builtins.open`
+            ``locale.getpreferredencoding(False)``, just as `open`
             does.
         prefer_file_temporary : `bool`, optional
             If `True`, for implementations that require transfers from a remote
@@ -1261,10 +1341,11 @@ class ResourcePath:
             may be necessary to avoid excessive memory usage by large files.
             Ignored by implementations that do not require a temporary.
 
-        Returns
-        -------
-        cm : `contextlib.ContextManager`
-            A context manager that wraps a file-like object.
+        Yields
+        ------
+        cm : `~contextlib.AbstractContextManager`
+            A context manager that wraps a `ResourceHandleProtocol` file-like
+            object.
 
         Notes
         -----
@@ -1277,7 +1358,7 @@ class ResourcePath:
         possible (as is guaranteed for local files).  `ResourcePath`
         implementations for which `as_local` does not return a temporary are
         required to reimplement `open`, though they may delegate to `super`
-        when `prefer_file_temporary` is `False`.
+        when ``prefer_file_temporary`` is `False`.
         """
         if self.dirLike:
             raise IsADirectoryError(f"Directory-like URI {self} cannot be opened.")
@@ -1304,9 +1385,7 @@ class ResourcePath:
                 yield handle
 
     @contextlib.contextmanager
-    def _openImpl(
-        self, mode: str = "r", *, encoding: Optional[str] = None
-    ) -> Iterator[ResourceHandleProtocol]:
+    def _openImpl(self, mode: str = "r", *, encoding: str | None = None) -> Iterator[ResourceHandleProtocol]:
         """Implement opening of a resource handle.
 
         This private method may be overridden by specific `ResourcePath`
@@ -1321,8 +1400,9 @@ class ResourcePath:
 
         Yields
         ------
-        handle : `BaseResourceHandle`
-            A handle that conforms to the `BaseResourcehandle interface
+        handle : `~._resourceHandles.BaseResourceHandle`
+            A handle that conforms to the
+            `~._resourceHandles.BaseResourceHandle` interface
 
         Notes
         -----
@@ -1331,10 +1411,7 @@ class ResourcePath:
         upon close. Subclasses of this class may offer more fine grained
         control.
         """
-        if "r" in mode or "a" in mode:
-            in_bytes = self.read()
-        else:
-            in_bytes = b""
+        in_bytes = self.read() if "r" in mode or "a" in mode else b""
         if "b" in mode:
             bytes_buffer = io.BytesIO(in_bytes)
             if "a" in mode:
@@ -1351,3 +1428,40 @@ class ResourcePath:
             out_bytes = str_buffer.getvalue().encode(encoding)
         if "r" not in mode or "+" in mode:
             self.write(out_bytes, overwrite=("x" not in mode))
+
+    def generate_presigned_get_url(self, *, expiration_time_seconds: int) -> str:
+        """Return a pre-signed URL that can be used to retrieve this resource
+        using an HTTP GET without supplying any access credentials.
+
+        Parameters
+        ----------
+        expiration_time_seconds : `int`
+            Number of seconds until the generated URL is no longer valid.
+
+        Returns
+        -------
+        url : `str`
+            HTTP URL signed for GET.
+        """
+        raise NotImplementedError(f"URL signing is not supported for '{self.scheme}'")
+
+    def generate_presigned_put_url(self, *, expiration_time_seconds: int) -> str:
+        """Return a pre-signed URL that can be used to upload a file to this
+        path using an HTTP PUT without supplying any access credentials.
+
+        Parameters
+        ----------
+        expiration_time_seconds : `int`
+            Number of seconds until the generated URL is no longer valid.
+
+        Returns
+        -------
+        url : `str`
+            HTTP URL signed for PUT.
+        """
+        raise NotImplementedError(f"URL signing is not supported for '{self.scheme}'")
+
+
+ResourcePathExpression = str | urllib.parse.ParseResult | ResourcePath | Path
+"""Type-annotation alias for objects that can be coerced to ResourcePath.
+"""
