@@ -11,6 +11,8 @@
 
 from __future__ import annotations
 
+__all__ = ("FileResourcePath",)
+
 import contextlib
 import copy
 import logging
@@ -20,14 +22,12 @@ import posixpath
 import re
 import shutil
 import urllib.parse
-
-__all__ = ("FileResourcePath",)
-
-from typing import IO, TYPE_CHECKING, Iterator, List, Optional, Tuple, Union
+from collections.abc import Iterator
+from typing import IO, TYPE_CHECKING
 
 from ._resourceHandles._fileResourceHandle import FileResourceHandle
 from ._resourcePath import ResourcePath
-from .utils import NoTransaction, os2posix, posix2os
+from .utils import NoTransaction, ensure_directory_is_writeable, os2posix, posix2os
 
 if TYPE_CHECKING:
     from .utils import TransactionProtocol
@@ -72,7 +72,7 @@ class FileResourcePath(ResourcePath):
         """Remove the resource."""
         os.remove(self.ospath)
 
-    def _as_local(self) -> Tuple[str, bool]:
+    def _as_local(self) -> tuple[str, bool]:
         """Return the local path of the file.
 
         This is an internal helper for ``as_local()``.
@@ -87,19 +87,14 @@ class FileResourcePath(ResourcePath):
         return self.ospath, self.isTemporary
 
     def read(self, size: int = -1) -> bytes:
-        """Return the entire content of the file as bytes."""
         with open(self.ospath, "rb") as fh:
             return fh.read(size)
 
     def write(self, data: bytes, overwrite: bool = True) -> None:
-        """Write the supplied data to the file."""
         dir = os.path.dirname(self.ospath)
         if not os.path.exists(dir):
-            os.makedirs(dir, exist_ok=True)
-        if overwrite:
-            mode = "wb"
-        else:
-            mode = "xb"
+            _create_directories(dir)
+        mode = "wb" if overwrite else "xb"
         with open(self.ospath, mode) as f:
             f.write(data)
 
@@ -115,7 +110,7 @@ class FileResourcePath(ResourcePath):
             Raised if a non-directory already exists.
         """
         try:
-            os.makedirs(self.ospath, exist_ok=True)
+            _create_directories(self.ospath)
         except FileExistsError:
             raise NotADirectoryError(f"{self.ospath} exists but is not a directory.") from None
 
@@ -135,7 +130,7 @@ class FileResourcePath(ResourcePath):
         src: ResourcePath,
         transfer: str,
         overwrite: bool = False,
-        transaction: Optional[TransactionProtocol] = None,
+        transaction: TransactionProtocol | None = None,
     ) -> None:
         """Transfer the current resource to a local file.
 
@@ -226,7 +221,7 @@ class FileResourcePath(ResourcePath):
             # The output location should not exist unless overwrite=True.
             # Rather than use `exists()`, use os.stat since we might need
             # the full answer later.
-            dest_stat: Optional[os.stat_result]
+            dest_stat: os.stat_result | None
             try:
                 # Do not read through links of the file itself.
                 dest_stat = os.lstat(self.ospath)
@@ -266,7 +261,7 @@ class FileResourcePath(ResourcePath):
                 # Must create the directory -- this can not be rolled back
                 # since another transfer running concurrently may
                 # be relying on this existing.
-                os.makedirs(outputDir, exist_ok=True)
+                _create_directories(outputDir)
 
             if transaction is None:
                 # Use a no-op transaction to reduce code duplication
@@ -276,13 +271,11 @@ class FileResourcePath(ResourcePath):
             # exist we have to remove it before we do the actual "transfer"
             # below
             if "link" in transfer and overwrite and dest_stat:
-                try:
-                    self.remove()
-                except Exception:
+                with contextlib.suppress(Exception):
                     # If this fails we ignore it since it's a problem
                     # that will manifest immediately below with a more relevant
                     # error message
-                    pass
+                    self.remove()
 
             if transfer == "move":
                 # If a rename works we try that since that is guaranteed to
@@ -342,7 +335,7 @@ class FileResourcePath(ResourcePath):
                 with transaction.undoWith(f"relsymlink to {local_src}", os.remove, newFullPath):
                     os.symlink(relPath, newFullPath)
             else:
-                raise NotImplementedError("Transfer type '{}' not supported.".format(transfer))
+                raise NotImplementedError(f"Transfer type '{transfer}' not supported.")
 
             # This was an explicit move requested from a remote resource
             # try to remove that remote resource. We check is_temporary because
@@ -352,8 +345,8 @@ class FileResourcePath(ResourcePath):
                 src.remove()
 
     def walk(
-        self, file_filter: Optional[Union[str, re.Pattern]] = None
-    ) -> Iterator[Union[List, Tuple[ResourcePath, List[str], List[str]]]]:
+        self, file_filter: str | re.Pattern | None = None
+    ) -> Iterator[list | tuple[ResourcePath, list[str], list[str]]]:
         """Walk the directory tree returning matching files and directories.
 
         Parameters
@@ -386,21 +379,20 @@ class FileResourcePath(ResourcePath):
     def _fixupPathUri(
         cls,
         parsed: urllib.parse.ParseResult,
-        root: Optional[Union[str, ResourcePath]] = None,
+        root: ResourcePath | None = None,
         forceAbsolute: bool = False,
         forceDirectory: bool = False,
-    ) -> Tuple[urllib.parse.ParseResult, bool]:
+    ) -> tuple[urllib.parse.ParseResult, bool]:
         """Fix up relative paths in URI instances.
 
         Parameters
         ----------
         parsed : `~urllib.parse.ParseResult`
             The result from parsing a URI using `urllib.parse`.
-        root : `str` or `ResourcePath`, optional
+        root : `ResourcePath`, optional
             Path to use as root when converting relative to absolute.
-            If `None`, it will be the current working directory. This
-            is a local file system path, or a file URI.  It is only used if
-            a file-scheme is used incorrectly with a relative path.
+            If `None`, it will be the current working directory. It is only
+            used if a file-scheme is used incorrectly with a relative path.
         forceAbsolute : `bool`, ignored
             Has no effect for this subclass. ``file`` URIs are always
             absolute.
@@ -455,13 +447,13 @@ class FileResourcePath(ResourcePath):
         replacements = {}
 
         if root is None:
-            root = os.path.abspath(os.path.curdir)
-        elif isinstance(root, ResourcePath):
+            root_str = os.path.abspath(os.path.curdir)
+        else:
             if root.scheme and root.scheme != "file":
                 raise RuntimeError(f"The override root must be a file URI not {root.scheme}")
-            root = os.path.abspath(root.ospath)
+            root_str = os.path.abspath(root.ospath)
 
-        replacements["path"] = posixpath.normpath(posixpath.join(os2posix(root), parsed.path))
+        replacements["path"] = posixpath.normpath(posixpath.join(os2posix(root_str), parsed.path))
 
         # normpath strips trailing "/" so put it back if necessary
         # Acknowledge that trailing separator exists.
@@ -482,7 +474,56 @@ class FileResourcePath(ResourcePath):
         self,
         mode: str = "r",
         *,
-        encoding: Optional[str] = None,
+        encoding: str | None = None,
     ) -> Iterator[IO]:
         with FileResourceHandle(mode=mode, log=log, filename=self.ospath, encoding=encoding) as buffer:
             yield buffer  # type: ignore
+
+
+def _create_directories(name: str | bytes) -> None:
+    """Create a directory and all of its parent directories that don't yet
+    exist.
+
+    Parameters
+    ----------
+    name : `str` or `bytes`
+        Path to the directory to be created
+
+    Notes
+    -----
+    The code in this function is duplicated from the Python standard library
+    function os.makedirs with one change: if the user has set a process umask
+    that prevents us from creating/accessing files in the newly created
+    directories, the permissions of the directories are altered to allow
+    owner-write and owner-traverse so that they can be used.
+    """
+    # These are optional parameters in the original function, but they can be
+    # constant here.
+    mode = 0o777
+    exist_ok = True
+
+    head, tail = os.path.split(name)
+    if not tail:
+        head, tail = os.path.split(head)
+    if head and tail and not os.path.exists(head):
+        try:
+            _create_directories(head)
+        except FileExistsError:
+            # Defeats race condition when another thread created the path
+            pass
+        cdir: str | bytes = os.curdir
+        if isinstance(tail, bytes):
+            cdir = bytes(os.curdir, "ASCII")
+        if tail == cdir:  # xxx/newdir/. exists if xxx/newdir exists
+            return
+    try:
+        os.mkdir(name, mode)
+        # This is the portion that is modified relative to the standard library
+        # version of the function.
+        ensure_directory_is_writeable(name)
+        # end modified portion
+    except OSError:
+        # Cannot rely on checking for EEXIST, since the operating system
+        # could give priority to other errors like EACCES or EROFS
+        if not exist_ok or not os.path.isdir(name):
+            raise
