@@ -19,7 +19,8 @@ import contextlib
 import logging
 import re
 import tempfile
-from typing import TYPE_CHECKING, Iterator, List, Optional, Set, Tuple, Union
+from collections.abc import Iterator
+from typing import TYPE_CHECKING
 
 from ._resourceHandles._baseResourceHandle import ResourceHandleProtocol
 
@@ -39,21 +40,33 @@ except ImportError:
 
     # Must also fake the exception classes.
     class ClientError(Exception):
+        """Generic client error."""
+
         pass
 
     class NotFound(ClientError):  # type: ignore  # noqa: N818
+        """Resource not found error."""
+
         pass
 
     class TooManyRequests(ClientError):  # type: ignore  # noqa: N818
+        """Too many requests error."""
+
         pass
 
     class InternalServerError(ClientError):  # type: ignore
+        """Internal server error."""
+
         pass
 
     class BadGateway(ClientError):  # type: ignore  # noqa: N818
+        """Bad gateway error."""
+
         pass
 
     class ServiceUnavailable(ClientError):  # type: ignore  # noqa: N818
+        """Service unavailable error."""
+
         pass
 
 
@@ -76,13 +89,23 @@ _RETRIEVABLE_TYPES = (
 
 
 def is_retryable(exc: Exception) -> bool:
+    """Report if the given exception is a condition that can be retried.
+
+    Parameters
+    ----------
+    exc : `Exception`
+        Exception to check.
+
+    Returns
+    -------
+    `bool`
+        Returns `True` if the given exception is a condition that can be
+        retried.
+    """
     return isinstance(exc, _RETRIEVABLE_TYPES)
 
 
-if retry:
-    _RETRY_POLICY = retry.Retry(predicate=is_retryable)
-else:
-    _RETRY_POLICY = None
+_RETRY_POLICY = retry.Retry(predicate=is_retryable) if retry else None
 
 
 _client = None
@@ -101,9 +124,9 @@ def _get_client() -> storage.Client:
 class GSResourcePath(ResourcePath):
     """Access Google Cloud Storage resources."""
 
-    _bucket: Optional[storage.Bucket] = None
-    _blob: Optional[storage.Blob] = None
-    _client: Optional[storage.Client] = None
+    _bucket: storage.Bucket | None = None
+    _blob: storage.Blob | None = None
+    _client: storage.Client | None = None
 
     @property
     def client(self) -> storage.Client:
@@ -134,7 +157,7 @@ class GSResourcePath(ResourcePath):
         try:
             self.blob.reload(retry=_RETRY_POLICY)
         except NotFound:
-            raise FileNotFoundError(f"Resource {self} does not exist")
+            raise FileNotFoundError(f"Resource {self} does not exist") from None
         size = self.blob.size
         if size is None:
             raise FileNotFoundError(f"Resource {self} does not exist")
@@ -161,9 +184,8 @@ class GSResourcePath(ResourcePath):
         return body
 
     def write(self, data: bytes, overwrite: bool = True) -> None:
-        if not overwrite:
-            if self.exists():
-                raise FileExistsError(f"Remote resource {self} exists and overwrite has been disabled")
+        if not overwrite and self.exists():
+            raise FileExistsError(f"Remote resource {self} exists and overwrite has been disabled")
         with time_this(log, msg="Write to %s", args=(self,)):
             self.blob.upload_from_string(data, retry=_RETRY_POLICY)
 
@@ -181,13 +203,15 @@ class GSResourcePath(ResourcePath):
         # Should this method do anything at all?
         self.blob.upload_from_string(b"", retry=_RETRY_POLICY)
 
-    def _as_local(self) -> Tuple[str, bool]:
-        with tempfile.NamedTemporaryFile(suffix=self.getExtension(), delete=False) as tmpFile:
-            with time_this(log, msg="Downloading %s to local file", args=(self,)):
-                try:
-                    self.blob.download_to_filename(tmpFile.name, retry=_RETRY_POLICY)
-                except NotFound as e:
-                    raise FileNotFoundError(f"No such resource: {self}") from e
+    def _as_local(self) -> tuple[str, bool]:
+        with (
+            tempfile.NamedTemporaryFile(suffix=self.getExtension(), delete=False) as tmpFile,
+            time_this(log, msg="Downloading %s to local file", args=(self,)),
+        ):
+            try:
+                self.blob.download_to_file(tmpFile, retry=_RETRY_POLICY)
+            except NotFound as e:
+                raise FileNotFoundError(f"No such resource: {self}") from e
         return tmpFile.name, True
 
     def transfer_from(
@@ -195,7 +219,7 @@ class GSResourcePath(ResourcePath):
         src: ResourcePath,
         transfer: str = "copy",
         overwrite: bool = False,
-        transaction: Optional[TransactionProtocol] = None,
+        transaction: TransactionProtocol | None = None,
     ) -> None:
         if transfer not in self.transferModes:
             raise ValueError(f"Transfer mode '{transfer}' not supported by URI scheme {self.scheme}")
@@ -247,9 +271,8 @@ class GSResourcePath(ResourcePath):
                         break
         else:
             # Use local file and upload it
-            with src.as_local() as local_uri:
-                with time_this(log, msg=timer_msg, args=timer_args):
-                    self.blob.upload_from_filename(local_uri.ospath, retry=_RETRY_POLICY)
+            with src.as_local() as local_uri, time_this(log, msg=timer_msg, args=timer_args):
+                self.blob.upload_from_filename(local_uri.ospath, retry=_RETRY_POLICY)
 
         # This was an explicit move requested from a remote resource
         # try to remove that resource
@@ -262,7 +285,7 @@ class GSResourcePath(ResourcePath):
         self,
         mode: str = "r",
         *,
-        encoding: Optional[str] = None,
+        encoding: str | None = None,
         prefer_file_temporary: bool = False,
     ) -> Iterator[ResourceHandleProtocol]:
         # Docstring inherited
@@ -291,8 +314,8 @@ class GSResourcePath(ResourcePath):
                 yield buffer
 
     def walk(
-        self, file_filter: Optional[Union[str, re.Pattern]] = None
-    ) -> Iterator[Union[List, Tuple[ResourcePath, List[str], List[str]]]]:
+        self, file_filter: str | re.Pattern | None = None
+    ) -> Iterator[list | tuple[ResourcePath, list[str], list[str]]]:
         # We pretend that GCS uses directories and files and not simply keys.
         if not (self.isdir() or self.is_root):
             raise ValueError(f"Can not walk a non-directory URI: {self}")
@@ -311,7 +334,7 @@ class GSResourcePath(ResourcePath):
         # them all grouped properly across the 1000 limit boundary.
         prefix = self.relativeToPathRoot if not self.is_root else ""
         prefix_len = len(prefix)
-        dirnames: Set[str] = set()
+        dirnames: set[str] = set()
         filenames = []
         files_there = False
 
@@ -335,7 +358,7 @@ class GSResourcePath(ResourcePath):
             # Nothing found so match os.walk and return immediately.
             return
         else:
-            yield self, list(sorted(dirnames)), filenames
+            yield self, sorted(dirnames), filenames
 
         for dir in sorted(dirnames):
             new_uri = self.join(dir)
