@@ -17,14 +17,22 @@ import contextlib
 import functools
 import io
 import logging
+import math
 import os
 import os.path
 import random
 import re
 import stat
 import tempfile
-import xml.etree.ElementTree as eTree
-from typing import TYPE_CHECKING, BinaryIO, Iterator, List, Optional, Tuple, Union, cast
+from collections.abc import Iterator
+from typing import TYPE_CHECKING, BinaryIO, cast
+
+try:
+    # Prefer 'defusedxml' (not part of standard library) if available, since
+    # 'xml' is vulnerable to XML bombs.
+    import defusedxml.ElementTree as eTree
+except ImportError:
+    import xml.etree.ElementTree as eTree
 
 import requests
 from astropy import units as u
@@ -43,17 +51,35 @@ if TYPE_CHECKING:
 log = logging.getLogger(__name__)
 
 
-# Default timeouts for all HTTP requests, in seconds.
-DEFAULT_TIMEOUT_CONNECT = 300
-DEFAULT_TIMEOUT_READ = 1500
+def _timeout_from_environment(env_var: str, default_value: float) -> float:
+    """Convert and return a timeout from the value of an environment variable
+    or a default value if the environment variable is not initialized. The
+    value of `env_var` must be a valid `float` otherwise this function raises.
 
-# Default number of connections to persist with both the front end and back end
-# servers.
-DEFAULT_FRONTEND_PERSISTENT_CONNECTIONS = "2"
-DEFAULT_BACKEND_PERSISTENT_CONNECTIONS = "1"
+    Parameters
+    ----------
+    env_var : `str`
+        Environment variable to look for.
+    default_value : `float``
+        Value to return if `env_var` is not defined in the environment.
 
-# Accepted digest algorithms
-ACCEPTED_DIGESTS = ("adler32", "md5", "sha-256", "sha-512")
+    Returns
+    -------
+    _timeout_from_environment : `float`
+        Converted value.
+    """
+    try:
+        timeout = float(os.environ.get(env_var, default_value))
+    except ValueError:
+        raise ValueError(
+            f"Expecting valid timeout value in environment variable {env_var} but found "
+            f"{os.environ.get(env_var)}"
+        ) from None
+
+    if math.isnan(timeout):
+        raise ValueError(f"Unexpected timeout value NaN found in environment variable {env_var}")
+
+    return timeout
 
 
 class HttpResourcePathConfig:
@@ -61,36 +87,65 @@ class HttpResourcePathConfig:
     HttpResourcePath.
     """
 
-    _front_end_connections: Optional[int] = None
-    _back_end_connections: Optional[int] = None
-    _digest_algorithm: Optional[str] = None
-    _send_expect_on_put: Optional[bool] = None
-    _timeout: Optional[tuple[int, int]] = None
+    # Default timeouts for all HTTP requests (seconds).
+    DEFAULT_TIMEOUT_CONNECT = 30.0
+    DEFAULT_TIMEOUT_READ = 1_500.0
+
+    # Default lower and upper bounds for the backoff interval (seconds).
+    # A value in this interval is randomly selected as the backoff factor when
+    # requests need to be retried.
+    DEFAULT_BACKOFF_MIN = 1.0
+    DEFAULT_BACKOFF_MAX = 3.0
+
+    # Default number of connections to persist with both the front end and
+    # back end servers.
+    DEFAULT_FRONTEND_PERSISTENT_CONNECTIONS = 2
+    DEFAULT_BACKEND_PERSISTENT_CONNECTIONS = 1
+
+    # Accepted digest algorithms
+    ACCEPTED_DIGESTS = ("adler32", "md5", "sha-256", "sha-512")
+
+    _front_end_connections: int | None = None
+    _back_end_connections: int | None = None
+    _digest_algorithm: str | None = None
+    _send_expect_on_put: bool | None = None
+    _timeout: tuple[float, float] | None = None
+    _collect_memory_usage: bool | None = None
+    _backoff_min: float | None = None
+    _backoff_max: float | None = None
 
     @property
     def front_end_connections(self) -> int:
         """Number of persistent connections to the front end server."""
-
         if self._front_end_connections is not None:
             return self._front_end_connections
 
-        self._front_end_connections = int(
-            os.environ.get(
-                "LSST_HTTP_FRONTEND_PERSISTENT_CONNECTIONS", DEFAULT_FRONTEND_PERSISTENT_CONNECTIONS
+        try:
+            self._front_end_connections = int(
+                os.environ.get(
+                    "LSST_HTTP_FRONTEND_PERSISTENT_CONNECTIONS", self.DEFAULT_FRONTEND_PERSISTENT_CONNECTIONS
+                )
             )
-        )
+        except ValueError:
+            self._front_end_connections = self.DEFAULT_FRONTEND_PERSISTENT_CONNECTIONS
+
         return self._front_end_connections
 
     @property
     def back_end_connections(self) -> int:
         """Number of persistent connections to the back end servers."""
-
         if self._back_end_connections is not None:
             return self._back_end_connections
 
-        self._back_end_connections = int(
-            os.environ.get("LSST_HTTP_BACKEND_PERSISTENT_CONNECTIONS", DEFAULT_BACKEND_PERSISTENT_CONNECTIONS)
-        )
+        try:
+            self._back_end_connections = int(
+                os.environ.get(
+                    "LSST_HTTP_BACKEND_PERSISTENT_CONNECTIONS", self.DEFAULT_BACKEND_PERSISTENT_CONNECTIONS
+                )
+            )
+        except ValueError:
+            self._back_end_connections = self.DEFAULT_BACKEND_PERSISTENT_CONNECTIONS
+
         return self._back_end_connections
 
     @property
@@ -104,12 +159,11 @@ class HttpResourcePathConfig:
             The name of a digest algorithm or the empty string if no algotihm
             is configured.
         """
-
         if self._digest_algorithm is not None:
             return self._digest_algorithm
 
         digest = os.environ.get("LSST_HTTP_DIGEST", "").lower()
-        if digest not in ACCEPTED_DIGESTS:
+        if digest not in self.ACCEPTED_DIGESTS:
             digest = ""
 
         self._digest_algorithm = digest
@@ -124,7 +178,6 @@ class HttpResourcePathConfig:
         the client knows how to handle redirects to the specific server that
         will actually receive the data for PUT requests.
         """
-
         if self._send_expect_on_put is not None:
             return self._send_expect_on_put
 
@@ -132,24 +185,71 @@ class HttpResourcePathConfig:
         return self._send_expect_on_put
 
     @property
-    def timeout(self) -> tuple[int, int]:
+    def timeout(self) -> tuple[float, float]:
         """Return a tuple with the values of timeouts for connecting to the
         server and reading its response, respectively. Both values are in
         seconds.
         """
-
         if self._timeout is not None:
             return self._timeout
 
         self._timeout = (
-            int(os.environ.get("LSST_HTTP_TIMEOUT_CONNECT", DEFAULT_TIMEOUT_CONNECT)),
-            int(os.environ.get("LSST_HTTP_TIMEOUT_READ", DEFAULT_TIMEOUT_READ)),
+            _timeout_from_environment("LSST_HTTP_TIMEOUT_CONNECT", self.DEFAULT_TIMEOUT_CONNECT),
+            _timeout_from_environment("LSST_HTTP_TIMEOUT_READ", self.DEFAULT_TIMEOUT_READ),
         )
         return self._timeout
 
+    @property
+    def collect_memory_usage(self) -> bool:
+        """Return true if we want to collect memory usage when timing
+        operations against the remote server via the `lsst.utils.time_this`
+        context manager.
+        """
+        if self._collect_memory_usage is not None:
+            return self._collect_memory_usage
+
+        self._collect_memory_usage = "LSST_HTTP_COLLECT_MEMORY_USAGE" in os.environ
+        return self._collect_memory_usage
+
+    @property
+    def backoff_min(self) -> float:
+        """Lower bound of the interval from which a backoff factor is randomly
+        selected when retrying requests (seconds).
+        """
+        if self._backoff_min is not None:
+            return self._backoff_min
+
+        self._backoff_min = self.DEFAULT_BACKOFF_MIN
+        try:
+            backoff_min = float(os.environ.get("LSST_HTTP_BACKOFF_MIN", self.DEFAULT_BACKOFF_MIN))
+            if not math.isnan(backoff_min):
+                self._backoff_min = backoff_min
+        except ValueError:
+            pass
+
+        return self._backoff_min
+
+    @property
+    def backoff_max(self) -> float:
+        """Upper bound of the interval from which a backoff factor is randomly
+        selected when retrying requests (seconds).
+        """
+        if self._backoff_max is not None:
+            return self._backoff_max
+
+        self._backoff_max = self.DEFAULT_BACKOFF_MAX
+        try:
+            backoff_max = float(os.environ.get("LSST_HTTP_BACKOFF_MAX", self.DEFAULT_BACKOFF_MAX))
+            if not math.isnan(backoff_max):
+                self._backoff_max = backoff_max
+        except ValueError:
+            pass
+
+        return self._backoff_max
+
 
 @functools.lru_cache
-def _is_webdav_endpoint(path: Union[ResourcePath, str]) -> bool:
+def _is_webdav_endpoint(path: ResourcePath | str) -> bool:
     """Check whether the remote HTTP endpoint implements WebDAV features.
 
     Parameters
@@ -165,41 +265,92 @@ def _is_webdav_endpoint(path: Union[ResourcePath, str]) -> bool:
         True if the endpoint implements WebDAV, False if it doesn't.
     """
     log.debug("Detecting HTTP endpoint type for '%s'...", path)
-    try:
-        ca_cert_bundle = os.getenv("LSST_HTTP_CACERT_BUNDLE")
-        verify: Union[bool, str] = ca_cert_bundle if ca_cert_bundle else True
-        resp = requests.options(str(path), verify=verify, stream=False)
 
-        # Check that "1" is part of the value of the "DAV" header. We don't
-        # use locks, so a server complying to class 1 is enough for our
-        # purposes. All webDAV servers must advertise at least compliance
-        # class "1".
-        #
-        # Compliance classes are documented in
-        #    http://www.webdav.org/specs/rfc4918.html#dav.compliance.classes
-        #
-        # Examples of values for header DAV are:
-        #   DAV: 1, 2
-        #   DAV: 1, <http://apache.org/dav/propset/fs/1>
-        if "DAV" not in resp.headers:
-            return False
-        else:
-            # Convert to str to keep mypy happy
-            compliance_class = str(resp.headers.get("DAV"))
-            return "1" in compliance_class.replace(" ", "").split(",")
+    # Send an OPTIONS request and inspect its response. An OPTIONS
+    # request does not need authentication of the client, so we don't need
+    # to provide a client certificate or a bearer token. We set a
+    # relatively short timeout since an OPTIONS request is relatively cheap
+    # for the server to compute.
+
+    # Create a session for configuring retries
+    retries = Retry(
+        # Total number of retries to allow. Takes precedence over other
+        # counts.
+        total=6,
+        # How many connection-related errors to retry on.
+        connect=3,
+        # How many times to retry on read errors.
+        read=3,
+        # How many times to retry on bad status codes.
+        status=5,
+        # Set of uppercased HTTP method verbs that we should retry on.
+        allowed_methods=frozenset(
+            [
+                "OPTIONS",
+            ]
+        ),
+        # HTTP status codes that we should force a retry on.
+        status_forcelist=frozenset(
+            [
+                requests.codes.too_many_requests,  # 429
+                requests.codes.internal_server_error,  # 500
+                requests.codes.bad_gateway,  # 502
+                requests.codes.service_unavailable,  # 503
+                requests.codes.gateway_timeout,  # 504
+            ]
+        ),
+        # Whether to respect 'Retry-After' header on status codes defined
+        # above.
+        respect_retry_after_header=True,
+    )
+
+    try:
+        session = requests.Session()
+        session.mount(str(path), HTTPAdapter(max_retries=retries))
+        session.verify = os.environ.get("LSST_HTTP_CACERT_BUNDLE", True)
+        with session:
+            resp = session.options(
+                str(path),
+                stream=False,
+                timeout=(
+                    _timeout_from_environment("LSST_HTTP_TIMEOUT_CONNECT", 30.0),
+                    _timeout_from_environment("LSST_HTTP_TIMEOUT_READ", 60.0),
+                ),
+            )
+            if resp.status_code not in (requests.codes.ok, requests.codes.created):
+                return False
+
+            # Check that "1" is part of the value of the "DAV" header. We don't
+            # use locks, so a server complying to class 1 is enough for our
+            # purposes. All webDAV servers must advertise at least compliance
+            # class "1".
+            #
+            # Compliance classes are documented in
+            #    http://www.webdav.org/specs/rfc4918.html#dav.compliance.classes
+            #
+            # Examples of values for header DAV are:
+            #   DAV: 1, 2
+            #   DAV: 1, <http://apache.org/dav/propset/fs/1>
+            if "DAV" not in resp.headers:
+                return False
+            else:
+                # Convert to str to keep mypy happy
+                compliance_class = str(resp.headers.get("DAV"))
+                return "1" in compliance_class.replace(" ", "").split(",")
+
     except requests.exceptions.SSLError as e:
         log.warning(
             "Environment variable LSST_HTTP_CACERT_BUNDLE can be used to "
-            "specify a bundle of certificate authorities you trust which are "
-            "not included in the default set of trusted authorities of your "
-            "system."
+            "specify tha path to a bundle of certificate authorities you trust "
+            "which are not included in the default set of trusted authorities "
+            "of this system."
         )
         raise e
 
 
 # Tuple (path, block_size) pointing to the location of a local directory
 # to save temporary files and the block size of the underlying file system.
-_TMPDIR: Optional[tuple[str, int]] = None
+_TMPDIR: tuple[str, int] | None = None
 
 
 def _get_temp_dir() -> tuple[str, int]:
@@ -267,35 +418,60 @@ class BearerTokenAuth(AuthBase):
                 self._token = f.read().rstrip("\n")
 
     def __call__(self, req: requests.PreparedRequest) -> requests.PreparedRequest:
-        if self._token:
+        # Only add a bearer token to a request when using secure HTTP.
+        if req.url and req.url.lower().startswith("https://") and self._token:
             self._refresh()
             req.headers["Authorization"] = f"Bearer {self._token}"
         return req
 
 
 class SessionStore:
-    """Cache a reusable HTTP client session per endpoint."""
+    """Cache a reusable HTTP client session per endpoint.
 
-    def __init__(self, num_pools: int = 10, max_persistent_connections: int = 1) -> None:
+    Parameters
+    ----------
+    num_pools : `int`, optional
+        Number of connection pools to keep: there is one pool per remote
+        host.
+    max_persistent_connections : `int`, optional
+        Maximum number of connections per remote host to persist in each
+        connection pool.
+    backoff_min : `float`, optional
+        Minimum value of the interval to compute the exponential
+        backoff factor when retrying requests (seconds).
+    backoff_max : `float`, optional
+        Maximum value of the interval to compute the exponential
+        backoff factor when retrying requests (seconds).
+    """
+
+    def __init__(
+        self,
+        num_pools: int = 10,
+        max_persistent_connections: int = 1,
+        backoff_min: float = 1.0,
+        backoff_max: float = 3.0,
+    ) -> None:
         # Dictionary to store the session associated to a given URI. The key
         # of the dictionary is a root URI and the value is the session.
         self._sessions: dict[str, requests.Session] = {}
 
-        # Number of connection pools to keep: there is one pool per remote
-        # host. See documentation of urllib3 PoolManager class:
+        # See documentation of urllib3 PoolManager class:
         # https://urllib3.readthedocs.io
-        self._num_pools = num_pools
+        self._num_pools: int = num_pools
 
-        # Maximum number of connections per remote host to persist in each
-        # connection pool. See urllib3 Advanced Usage documentation:
+        # See urllib3 Advanced Usage documentation:
         # https://urllib3.readthedocs.io/en/stable/advanced-usage.html
-        self._max_persistent_connections = max_persistent_connections
+        self._max_persistent_connections: int = max_persistent_connections
+
+        # Minimum and maximum values of the interval to compute the exponential
+        # backoff factor when retrying requests (seconds).
+        self._backoff_min: float = backoff_min
+        self._backoff_max: float = backoff_max if backoff_max > backoff_min else backoff_min + 1.0
 
     def clear(self) -> None:
         """Destroy all previously created sessions and attempt to close
         underlying idle network connections.
         """
-
         # Close all sessions and empty the store. Idle network connections
         # should be closed as a consequence. We don't have means through
         # the API exposed by Requests to actually force closing the
@@ -358,7 +534,6 @@ class SessionStore:
         session = requests.Session()
         root_uri = str(rpath.root_uri())
         log.debug("Creating new HTTP session for endpoint %s ...", root_uri)
-
         retries = Retry(
             # Total number of retries to allow. Takes precedence over other
             # counts.
@@ -368,8 +543,9 @@ class SessionStore:
             # How many times to retry on read errors.
             read=3,
             # Backoff factor to apply between attempts after the second try
-            # (seconds)
-            backoff_factor=30 * (1 + random.random()),
+            # (seconds). Compute a random jitter to prevent all the clients
+            # to overwhelm the server by sending requests at the same time.
+            backoff_factor=self._backoff_min + (self._backoff_max - self._backoff_min) * random.random(),
             # How many times to retry on bad status codes.
             status=5,
             # Set of uppercased HTTP method verbs that we should retry on.
@@ -511,7 +687,7 @@ class HttpResourcePath(ResourcePath):
         Valid values are those in ACCEPTED_DIGESTS.
     """
 
-    _is_webdav: Optional[bool] = None
+    _is_webdav: bool | None = None
 
     # Configuration items for this class instances.
     _config = HttpResourcePathConfig()
@@ -524,6 +700,8 @@ class HttpResourcePath(ResourcePath):
     _metadata_session_store = SessionStore(
         num_pools=5,
         max_persistent_connections=_config.front_end_connections,
+        backoff_min=_config.backoff_min,
+        backoff_max=_config.backoff_max,
     )
 
     # The data session is used for interaction with the front end servers which
@@ -536,9 +714,11 @@ class HttpResourcePath(ResourcePath):
     _data_session_store = SessionStore(
         num_pools=25,
         max_persistent_connections=_config.back_end_connections,
+        backoff_min=_config.backoff_min,
+        backoff_max=_config.backoff_max,
     )
 
-    # Process ID which created the sessions above. We need to store this
+    # Process ID which created the session stores above. We need to store this
     # to replace sessions created by a parent process and inherited by a
     # child process after a fork, to avoid confusing the SSL layer.
     _pid: int = -1
@@ -548,34 +728,39 @@ class HttpResourcePath(ResourcePath):
         """Client session to send requests which do not require upload or
         download of data, i.e. mostly metadata requests.
         """
+        if hasattr(self, "_metadata_session"):
+            if HttpResourcePath._pid == os.getpid():
+                return self._metadata_session
+            else:
+                # The metadata session we have in cache was likely created by
+                # a parent process. Discard all the sessions in that store.
+                self._metadata_session_store.clear()
 
-        if hasattr(self, "_metadata_session") and self._pid == os.getpid():
-            return self._metadata_session
-
-        # Reset the store in case it was created by another process and
-        # retrieve a session.
-        self._metadata_session_store.clear()
-        self._pid = os.getpid()
+        # Retrieve a new metadata session.
+        HttpResourcePath._pid = os.getpid()
         self._metadata_session: requests.Session = self._metadata_session_store.get(self)
         return self._metadata_session
 
     @property
     def data_session(self) -> requests.Session:
         """Client session for uploading and downloading data."""
+        if hasattr(self, "_data_session"):
+            if HttpResourcePath._pid == os.getpid():
+                return self._data_session
+            else:
+                # The data session we have in cache was likely created by
+                # a parent process. Discard all the sessions in that store.
+                self._data_session_store.clear()
 
-        if hasattr(self, "_data_session") and self._pid == os.getpid():
-            return self._data_session
-
-        # Reset the store in case it was created by another process and
-        # retrieve a session.
-        self._data_session_store.clear()
-        self._pid = os.getpid()
+        # Retrieve a new data session.
+        HttpResourcePath._pid = os.getpid()
         self._data_session: requests.Session = self._data_session_store.get(self)
         return self._data_session
 
     def _clear_sessions(self) -> None:
-        """Internal method to close the socket connections still open. Used
-        only in test suites to avoid warnings.
+        """Close the socket connections that are still open.
+
+        Used only in test suites to avoid warnings.
         """
         self._metadata_session_store.clear()
         self._data_session_store.clear()
@@ -714,12 +899,11 @@ class HttpResourcePath(ResourcePath):
             The number of bytes to read. Negative or omitted indicates
             that all data should be read.
         """
-
         # Use the data session as a context manager to ensure that the
         # network connections to both the front end and back end servers are
         # closed after downloading the data.
         log.debug("Reading from remote resource: %s", self.geturl())
-        stream = True if size > 0 else False
+        stream = size > 0
         with self.data_session as session:
             with time_this(log, msg="GET %s", args=(self,)):
                 resp = session.get(self.geturl(), stream=stream, timeout=self._config.timeout)
@@ -746,12 +930,15 @@ class HttpResourcePath(ResourcePath):
             the write will fail.
         """
         log.debug("Writing to remote resource: %s", self.geturl())
-        if not overwrite:
-            if self.exists():
-                raise FileExistsError(f"Remote resource {self} exists and overwrite has been disabled")
+        if not overwrite and self.exists():
+            raise FileExistsError(f"Remote resource {self} exists and overwrite has been disabled")
 
         # Ensure the parent directory exists.
-        self.parent().mkdir()
+        # This is only meaningful and appropriate for WebDAV, not the general
+        # HTTP case.  e.g. for S3 HTTP URLs, the underlying service has no
+        # concept of 'directories' at all.
+        if self.is_webdav_endpoint:
+            self.parent().mkdir()
 
         # Upload the data.
         log.debug("Writing data to remote resource: %s", self.geturl())
@@ -762,7 +949,7 @@ class HttpResourcePath(ResourcePath):
         src: ResourcePath,
         transfer: str = "copy",
         overwrite: bool = False,
-        transaction: Optional[TransactionProtocol] = None,
+        transaction: TransactionProtocol | None = None,
     ) -> None:
         """Transfer the current resource to a Webdav repository.
 
@@ -773,6 +960,8 @@ class HttpResourcePath(ResourcePath):
         transfer : `str`
             Mode to use for transferring the resource. Supports the following
             options: copy.
+        overwrite : `bool`, optional
+            Whether overwriting the remote resource is allowed or not.
         transaction : `~lsst.resources.utils.TransactionProtocol`, optional
             Currently unused.
         """
@@ -823,13 +1012,15 @@ class HttpResourcePath(ResourcePath):
             src.remove()
 
     def walk(
-        self, file_filter: Optional[Union[str, re.Pattern]] = None
-    ) -> Iterator[Union[List, Tuple[ResourcePath, List[str], List[str]]]]:
+        self, file_filter: str | re.Pattern | None = None
+    ) -> Iterator[list | tuple[ResourcePath, list[str], list[str]]]:
         """Walk the directory tree returning matching files and directories.
+
         Parameters
         ----------
         file_filter : `str` or `re.Pattern`, optional
             Regex to filter out files from the list before it is returned.
+
         Yields
         ------
         dirpath : `ResourcePath`
@@ -851,15 +1042,15 @@ class HttpResourcePath(ResourcePath):
 
         resp = self._propfind(depth="1")
         if resp.status_code == requests.codes.multi_status:  # 207
-            files: List[str] = []
-            dirs: List[str] = []
+            files: list[str] = []
+            dirs: list[str] = []
 
             for prop in _parse_propfind_response_body(resp.text):
                 if prop.is_file:
                     files.append(prop.name)
-                elif not self.path.endswith(prop.href):
-                    # Only include the names of sub-directories not the
-                    # directory being walked.
+                elif not prop.href.rstrip("/").endswith(self.path.rstrip("/")):
+                    # Only include the names of sub-directories not the name of
+                    # the directory being walked.
                     dirs.append(prop.name)
 
             if file_filter is not None:
@@ -874,7 +1065,7 @@ class HttpResourcePath(ResourcePath):
                 new_uri = self.join(dir, forceDirectory=True)
                 yield from new_uri.walk(file_filter)
 
-    def _as_local(self) -> Tuple[str, bool]:
+    def _as_local(self) -> tuple[str, bool]:
         """Download object over HTTP and place in temporary directory.
 
         Returns
@@ -884,7 +1075,6 @@ class HttpResourcePath(ResourcePath):
         temporary : `bool`
             Always returns `True`. This is always a temporary file.
         """
-
         # Use the session as a context manager to ensure that connections
         # to both the front end and back end servers are closed after the
         # download operation is finished.
@@ -895,26 +1085,31 @@ class HttpResourcePath(ResourcePath):
                     f"Unable to download resource {self}; status: {resp.status_code} {resp.reason}"
                 )
 
-            content_length = 0
-            expected_length = int(resp.headers.get("Content-Length", "-1"))
             tmpdir, buffering = _get_temp_dir()
-
             with tempfile.NamedTemporaryFile(
                 suffix=self.getExtension(), buffering=buffering, dir=tmpdir, delete=False
             ) as tmpFile:
+                expected_length = int(resp.headers.get("Content-Length", "-1"))
                 with time_this(
                     log,
                     msg="GET %s [length=%d] to local file %s [chunk_size=%d]",
                     args=(self, expected_length, tmpFile.name, buffering),
-                    mem_usage=True,
+                    mem_usage=self._config.collect_memory_usage,
                     mem_unit=u.mebibyte,
                 ):
+                    content_length = 0
                     for chunk in resp.iter_content(chunk_size=buffering):
                         tmpFile.write(chunk)
                         content_length += len(chunk)
 
-            # Check that the expected and actual content lengths match
-            if expected_length >= 0 and expected_length != content_length:
+            # Check that the expected and actual content lengths match. Perform
+            # this check only when the contents of the file was not encoded by
+            # the server.
+            if (
+                "Content-Encoding" not in resp.headers
+                and expected_length >= 0
+                and expected_length != content_length
+            ):
                 raise ValueError(
                     f"Size of downloaded file does not match value in Content-Length header for {self}: "
                     f"expecting {expected_length} and got {content_length} bytes"
@@ -925,11 +1120,11 @@ class HttpResourcePath(ResourcePath):
     def _send_webdav_request(
         self,
         method: str,
-        url: Optional[str] = None,
-        headers: dict[str, str] = {},
-        body: Optional[str] = None,
-        session: Optional[requests.Session] = None,
-        timeout: Optional[tuple[int, int]] = None,
+        url: str | None = None,
+        headers: dict[str, str] | None = None,
+        body: str | None = None,
+        session: requests.Session | None = None,
+        timeout: tuple[float, float] | None = None,
     ) -> requests.Response:
         """Send a webDAV request and correctly handle redirects.
 
@@ -940,7 +1135,7 @@ class HttpResourcePath(ResourcePath):
         headers : `dict`, optional
             A dictionary of key-value pairs (both strings) to include as
             headers in the request.
-        body: `str`, optional
+        body : `str`, optional
             The body of the request.
 
         Notes
@@ -969,6 +1164,9 @@ class HttpResourcePath(ResourcePath):
         if url is None:
             url = self.geturl()
 
+        if headers is None:
+            headers = {}
+
         if session is None:
             session = self.metadata_session
 
@@ -982,7 +1180,7 @@ class HttpResourcePath(ResourcePath):
                 method,
                 url,
             ),
-            mem_usage=True,
+            mem_usage=self._config.collect_memory_usage,
             mem_unit=u.mebibyte,
         ):
             for _ in range(max_redirects := 5):
@@ -1003,11 +1201,10 @@ class HttpResourcePath(ResourcePath):
             # We reached the maximum allowed number of redirects.
             # Stop trying.
             raise ValueError(
-                f"Could not get a response to {method} request for {self} after "
-                f"{max_redirects} redirections"
+                f"Could not get a response to {method} request for {self} after {max_redirects} redirections"
             )
 
-    def _propfind(self, body: Optional[str] = None, depth: str = "0") -> requests.Response:
+    def _propfind(self, body: str | None = None, depth: str = "0") -> requests.Response:
         """Send a PROPFIND webDAV request and return the response.
 
         Parameters
@@ -1054,12 +1251,16 @@ class HttpResourcePath(ResourcePath):
 
     def _options(self) -> requests.Response:
         """Send a OPTIONS webDAV request for this resource."""
+        resp = self._send_webdav_request("OPTIONS")
+        if resp.status_code in (requests.codes.ok, requests.codes.created):
+            return resp
 
-        return self._send_webdav_request("OPTIONS")
+        raise ValueError(
+            f"Unexpected response to OPTIONS request for {self}, status: {resp.status_code} {resp.reason}"
+        )
 
     def _head(self) -> requests.Response:
         """Send a HEAD webDAV request for this resource."""
-
         return self._send_webdav_request("HEAD")
 
     def _mkcol(self) -> None:
@@ -1078,7 +1279,6 @@ class HttpResourcePath(ResourcePath):
 
     def _delete(self) -> None:
         """Send a DELETE webDAV request for this resource."""
-
         log.debug("Deleting %s ...", self.geturl())
 
         # If this is a directory, ensure the remote is a webDAV server because
@@ -1138,7 +1338,6 @@ class HttpResourcePath(ResourcePath):
         method : `str`
             The method to perform. Valid values are "COPY" or "MOVE" (in
             uppercase).
-
         src : `HttpResourcePath`
             The source of the contents to move to `self`.
         """
@@ -1193,7 +1392,7 @@ class HttpResourcePath(ResourcePath):
         """
         return self._copy_or_move("MOVE", src)
 
-    def _put(self, data: Union[BinaryIO, bytes]) -> None:
+    def _put(self, data: BinaryIO | bytes) -> None:
         """Perform an HTTP PUT request and handle redirection.
 
         Parameters
@@ -1216,7 +1415,13 @@ class HttpResourcePath(ResourcePath):
             # Send an empty PUT request to get redirected to the final
             # destination.
             log.debug("Sending empty PUT request to %s", url)
-            with time_this(log, msg="PUT (no data) %s", args=(url,), mem_usage=True, mem_unit=u.mebibyte):
+            with time_this(
+                log,
+                msg="PUT (no data) %s",
+                args=(url,),
+                mem_usage=self._config.collect_memory_usage,
+                mem_unit=u.mebibyte,
+            ):
                 resp = session.request(
                     "PUT",
                     url,
@@ -1244,11 +1449,17 @@ class HttpResourcePath(ResourcePath):
             # for the list of supported digest algorithhms.
             # In addition, note that not all servers implement this RFC so
             # the checksum may not be computed by the server.
-            put_headers: Optional[dict[str, str]] = None
+            put_headers: dict[str, str] | None = None
             if digest := self._config.digest_algorithm:
                 put_headers = {"Want-Digest": digest}
 
-            with time_this(log, msg="PUT %s", args=(url,), mem_usage=True, mem_unit=u.mebibyte):
+            with time_this(
+                log,
+                msg="PUT %s",
+                args=(url,),
+                mem_usage=self._config.collect_memory_usage,
+                mem_unit=u.mebibyte,
+            ):
                 resp = session.request(
                     "PUT",
                     url,
@@ -1272,7 +1483,7 @@ class HttpResourcePath(ResourcePath):
         self,
         mode: str = "r",
         *,
-        encoding: Optional[str] = None,
+        encoding: str | None = None,
     ) -> Iterator[ResourceHandleProtocol]:
         resp = self._head()
         accepts_range = resp.status_code == requests.codes.ok and resp.headers.get("Accept-Ranges") == "bytes"
@@ -1345,7 +1556,7 @@ def _is_protected(filepath: str) -> bool:
     return owner_accessible and not group_accessible and not other_accessible
 
 
-def _parse_propfind_response_body(body: str) -> List[DavProperty]:
+def _parse_propfind_response_body(body: str) -> list[DavProperty]:
     """Parse the XML-encoded contents of the response body to a webDAV PROPFIND
     request.
 
@@ -1405,19 +1616,24 @@ def _parse_propfind_response_body(body: str) -> List[DavProperty]:
         return responses
     else:
         # Could not parse the body
-        raise ValueError(f"Unable to parse response for PROPFIND request: {response}")
+        raise ValueError(f"Unable to parse response for PROPFIND request: {body}")
 
 
 class DavProperty:
     """Helper class to encapsulate select live DAV properties of a single
     resource, as retrieved via a PROPFIND request.
+
+    Parameters
+    ----------
+    response : `eTree.Element` or `None`
+        The XML response defining the DAV property.
     """
 
     # Regular expression to compare against the 'status' element of a
     # PROPFIND response's 'propstat' element.
     _status_ok_rex = re.compile(r"^HTTP/.* 200 .*$", re.IGNORECASE)
 
-    def __init__(self, response: Optional[eTree.Element]):
+    def __init__(self, response: eTree.Element | None):
         self._href: str = ""
         self._displayname: str = ""
         self._collection: bool = False
@@ -1428,11 +1644,16 @@ class DavProperty:
             self._parse(response)
 
     def _parse(self, response: eTree.Element) -> None:
-        # Extract 'href'
+        # Extract 'href'.
         if (element := response.find("./{DAV:}href")) is not None:
             # We need to use "str(element.text)"" instead of "element.text" to
-            # keep mypy happy
+            # keep mypy happy.
             self._href = str(element.text).strip()
+        else:
+            raise ValueError(
+                "Property 'href' expected but not found in PROPFIND response: "
+                f"{eTree.tostring(response, encoding='unicode')}"
+            )
 
         for propstat in response.findall("./{DAV:}propstat"):
             # Only extract properties of interest with status OK.
@@ -1457,6 +1678,16 @@ class DavProperty:
                 if (element := prop.find("./{DAV:}displayname")) is not None:
                     self._displayname = str(element.text)
 
+        # Some webDAV servers don't include the 'displayname' property in the
+        # response so try to infer it from the value of the 'href' property.
+        # Depending on the server the href value may end with '/'.
+        if not self._displayname:
+            self._displayname = os.path.basename(self._href.rstrip("/"))
+
+        # Force a size of 0 for collections.
+        if self._collection:
+            self._getcontentlength = 0
+
     @property
     def exists(self) -> bool:
         # It is either a directory or a file with length of at least zero
@@ -1468,11 +1699,10 @@ class DavProperty:
 
     @property
     def is_file(self) -> bool:
-        return self._getcontentlength >= 0
+        return not self._collection
 
     @property
     def size(self) -> int:
-        # Only valid if is_file is True
         return self._getcontentlength
 
     @property
