@@ -11,6 +11,8 @@
 
 from __future__ import annotations
 
+__all__ = ("S3ResourcePath",)
+
 import contextlib
 import io
 import logging
@@ -18,10 +20,8 @@ import re
 import sys
 import tempfile
 import threading
-
-__all__ = ("S3ResourcePath",)
-
-from typing import IO, TYPE_CHECKING, Iterator, List, Optional, Tuple, Union, cast
+from collections.abc import Iterable, Iterator
+from typing import IO, TYPE_CHECKING, cast
 
 from botocore.exceptions import ClientError
 from lsst.utils.timer import time_this
@@ -30,7 +30,7 @@ from ._resourceHandles._baseResourceHandle import ResourceHandleProtocol
 from ._resourceHandles._s3ResourceHandle import S3ResourceHandle
 from ._resourcePath import ResourcePath
 from .s3utils import (
-    _TooManyRequestsException,
+    _TooManyRequestsError,
     all_retryable_errors,
     backoff,
     bucketExists,
@@ -41,10 +41,9 @@ from .s3utils import (
 )
 
 if TYPE_CHECKING:
-    try:
+    with contextlib.suppress(ImportError):
         import boto3
-    except ImportError:
-        pass
+
     from .utils import TransactionProtocol
 
 
@@ -52,12 +51,24 @@ log = logging.getLogger(__name__)
 
 
 class ProgressPercentage:
-    """Progress bar for S3 file uploads."""
+    """Progress bar for S3 file uploads.
+
+    Parameters
+    ----------
+    file : `ResourcePath`
+        Resource that is relevant to the progress percentage. The size of this
+        resource will be used to determine progress. The name will be used
+        in the log messages unless overridden by ``file_for_msg``.
+    file_for_msg : `ResourcePath` or `None`, optional
+        Resource name to include in log messages in preference to ``file``.
+    msg : `str`, optional
+        Message text to be included in every progress log message.
+    """
 
     log_level = logging.DEBUG
     """Default log level to use when issuing a message."""
 
-    def __init__(self, file: ResourcePath, file_for_msg: Optional[ResourcePath] = None, msg: str = ""):
+    def __init__(self, file: ResourcePath, file_for_msg: ResourcePath | None = None, msg: str = ""):
         self._filename = file
         self._file_for_msg = str(file_for_msg) if file_for_msg is not None else str(file)
         self._size = file.size()
@@ -91,14 +102,14 @@ def _translate_client_error(err: ClientError) -> None:
 
     Raises
     ------
-    _TooManyRequestsException
+    _TooManyRequestsError
         Raised if the `ClientError` looks like a 429 retry request.
     """
     if "(429)" in str(err):
         # ClientError includes the error code in the message
         # but no direct way to access it without looking inside the
         # response.
-        raise _TooManyRequestsException(str(err)) from err
+        raise _TooManyRequestsError(str(err)) from err
     elif "(404)" in str(err):
         # Some systems can generate this rather than NoSuchKey.
         raise FileNotFoundError("Resource not found: {self}")
@@ -113,6 +124,13 @@ class S3ResourcePath(ResourcePath):
         # Defer import for circular dependencies
         return getS3Client()
 
+    @classmethod
+    def _mexists(cls, uris: Iterable[ResourcePath]) -> dict[ResourcePath, bool]:
+        # Force client to be created before creating threads.
+        getS3Client()
+
+        return super()._mexists(uris)
+
     @backoff.on_exception(backoff.expo, retryable_io_errors, max_time=max_retry_time)
     def exists(self) -> bool:
         """Check that the S3 resource exists."""
@@ -146,7 +164,6 @@ class S3ResourcePath(ResourcePath):
 
     @backoff.on_exception(backoff.expo, all_retryable_errors, max_time=max_retry_time)
     def read(self, size: int = -1) -> bytes:
-        """Read the contents of the resource."""
         args = {}
         if size > 0:
             args["Range"] = f"bytes=0-{size-1}"
@@ -164,10 +181,8 @@ class S3ResourcePath(ResourcePath):
 
     @backoff.on_exception(backoff.expo, all_retryable_errors, max_time=max_retry_time)
     def write(self, data: bytes, overwrite: bool = True) -> None:
-        """Write the supplied data to the resource."""
-        if not overwrite:
-            if self.exists():
-                raise FileExistsError(f"Remote resource {self} exists and overwrite has been disabled")
+        if not overwrite and self.exists():
+            raise FileExistsError(f"Remote resource {self} exists and overwrite has been disabled")
         with time_this(log, msg="Write to %s", args=(self,)):
             self.client.put_object(Bucket=self.netloc, Key=self.relativeToPathRoot, Body=data)
 
@@ -181,11 +196,11 @@ class S3ResourcePath(ResourcePath):
             raise NotADirectoryError(f"Can not create a 'directory' for file-like URI {self}")
 
         # don't create S3 key when root is at the top-level of an Bucket
-        if not self.path == "/":
+        if self.path != "/":
             self.client.put_object(Bucket=self.netloc, Key=self.relativeToPathRoot)
 
     @backoff.on_exception(backoff.expo, all_retryable_errors, max_time=max_retry_time)
-    def _download_file(self, local_file: IO, progress: Optional[ProgressPercentage]) -> None:
+    def _download_file(self, local_file: IO, progress: ProgressPercentage | None) -> None:
         """Download the remote resource to a local file.
 
         Helper routine for _as_local to allow backoff without regenerating
@@ -202,7 +217,7 @@ class S3ResourcePath(ResourcePath):
             _translate_client_error(err)
             raise
 
-    def _as_local(self) -> Tuple[str, bool]:
+    def _as_local(self) -> tuple[str, bool]:
         """Download object from S3 and place in temporary directory.
 
         Returns
@@ -212,18 +227,20 @@ class S3ResourcePath(ResourcePath):
         temporary : `bool`
             Always returns `True`. This is always a temporary file.
         """
-        with tempfile.NamedTemporaryFile(suffix=self.getExtension(), delete=False) as tmpFile:
-            with time_this(log, msg="Downloading %s to local file", args=(self,)):
-                progress = (
-                    ProgressPercentage(self, msg="Downloading:")
-                    if log.isEnabledFor(ProgressPercentage.log_level)
-                    else None
-                )
-                self._download_file(tmpFile, progress)
+        with (
+            tempfile.NamedTemporaryFile(suffix=self.getExtension(), delete=False) as tmpFile,
+            time_this(log, msg="Downloading %s to local file", args=(self,)),
+        ):
+            progress = (
+                ProgressPercentage(self, msg="Downloading:")
+                if log.isEnabledFor(ProgressPercentage.log_level)
+                else None
+            )
+            self._download_file(tmpFile, progress)
         return tmpFile.name, True
 
     @backoff.on_exception(backoff.expo, all_retryable_errors, max_time=max_retry_time)
-    def _upload_file(self, local_file: ResourcePath, progress: Optional[ProgressPercentage]) -> None:
+    def _upload_file(self, local_file: ResourcePath, progress: ProgressPercentage | None) -> None:
         """Upload a local file with backoff.
 
         Helper method to wrap file uploading in backoff for transfer_from.
@@ -257,7 +274,7 @@ class S3ResourcePath(ResourcePath):
         src: ResourcePath,
         transfer: str = "copy",
         overwrite: bool = False,
-        transaction: Optional[TransactionProtocol] = None,
+        transaction: TransactionProtocol | None = None,
     ) -> None:
         """Transfer the current resource to an S3 bucket.
 
@@ -333,8 +350,8 @@ class S3ResourcePath(ResourcePath):
 
     @backoff.on_exception(backoff.expo, all_retryable_errors, max_time=max_retry_time)
     def walk(
-        self, file_filter: Optional[Union[str, re.Pattern]] = None
-    ) -> Iterator[Union[List, Tuple[ResourcePath, List[str], List[str]]]]:
+        self, file_filter: str | re.Pattern | None = None
+    ) -> Iterator[list | tuple[ResourcePath, list[str], list[str]]]:
         """Walk the directory tree returning matching files and directories.
 
         Parameters
@@ -411,7 +428,7 @@ class S3ResourcePath(ResourcePath):
         self,
         mode: str = "r",
         *,
-        encoding: Optional[str] = None,
+        encoding: str | None = None,
     ) -> Iterator[ResourceHandleProtocol]:
         with S3ResourceHandle(mode, log, self.client, self.netloc, self.relativeToPathRoot) as handle:
             if "b" in mode:
@@ -423,3 +440,18 @@ class S3ResourcePath(ResourcePath):
                 # BytesIO in the inheritance tree
                 with io.TextIOWrapper(cast(io.BytesIO, handle), encoding=encoding, write_through=True) as sub:
                     yield sub
+
+    def generate_presigned_get_url(self, *, expiration_time_seconds: int) -> str:
+        # Docstring inherited
+        return self._generate_presigned_url("get_object", expiration_time_seconds)
+
+    def generate_presigned_put_url(self, *, expiration_time_seconds: int) -> str:
+        # Docstring inherited
+        return self._generate_presigned_url("put_object", expiration_time_seconds)
+
+    def _generate_presigned_url(self, method: str, expiration_time_seconds: int) -> str:
+        return self.client.generate_presigned_url(
+            method,
+            Params={"Bucket": self.netloc, "Key": self.relativeToPathRoot},
+            ExpiresIn=expiration_time_seconds,
+        )
