@@ -20,7 +20,8 @@ import string
 import unittest
 import urllib.parse
 import uuid
-from typing import Any, Callable, Iterable, Optional, Union
+from collections.abc import Iterable
+from typing import TYPE_CHECKING, Any
 
 from lsst.resources import ResourcePath
 from lsst.resources.utils import makeTestTempDir, removeTestTempDir
@@ -29,7 +30,7 @@ TESTDIR = os.path.abspath(os.path.dirname(__file__))
 
 
 def _check_open(
-    test_case: Union[_GenericTestCase, unittest.TestCase],
+    test_case: _GenericTestCase | unittest.TestCase,
     uri: ResourcePath,
     *,
     mode_suffixes: Iterable[str] = ("", "t", "b"),
@@ -45,14 +46,14 @@ def _check_open(
         URI to use for tests.  Must point to a writeable location that is not
         yet occupied by a file.  On return, the location may point to a file
         only if the test fails.
-    mode_suffixes : `Iterable` of `str`
+    mode_suffixes : `~collections.abc.Iterable` of `str`
         Suffixes to pass as part of the ``mode`` argument to
         `ResourcePath.open`, indicating whether to open as binary or as text;
         the only permitted elements are ``""``, ``"t"``, and ``"b"`.
     **kwargs
         Additional keyword arguments to forward to all calls to `open`.
     """
-    text_content = "wxyz🙂"
+    text_content = "abcdefghijklmnopqrstuvwxyz🙂"
     bytes_content = uuid.uuid4().bytes
     content_by_mode_suffix = {
         "": text_content,
@@ -84,6 +85,55 @@ def _check_open(
         # Read the file we created and check the contents.
         with uri.open("r" + mode_suffix, **kwargs) as read_buffer:
             test_case.assertEqual(read_buffer.read(), content)
+        # Check that we can read bytes in a loop and get EOF
+        with uri.open("r" + mode_suffix, **kwargs) as read_buffer:
+            # Seek off the end of the file and should read empty back.
+            read_buffer.seek(1024)
+            test_case.assertEqual(read_buffer.tell(), 1024)
+            content_read = read_buffer.read()  # Read as much as we can.
+            test_case.assertEqual(len(content_read), 0, f"Read: {content_read!r}, expected empty.")
+
+            # First read more than the content.
+            read_buffer.seek(0)
+            size = len(content) * 3
+            chunk_read = read_buffer.read(size)
+            test_case.assertEqual(chunk_read, content)
+
+            # Repeated reads should always return empty string.
+            chunk_read = read_buffer.read(size)
+            test_case.assertEqual(len(chunk_read), 0)
+            chunk_read = read_buffer.read(size)
+            test_case.assertEqual(len(chunk_read), 0)
+
+            # Go back to start of file and read in smaller chunks.
+            read_buffer.seek(0)
+            size = len(content) // 3
+
+            content_read = empty_content_by_mode_suffix[mode_suffix]
+            n_reads = 0
+            while chunk_read := read_buffer.read(size):
+                content_read += chunk_read
+                n_reads += 1
+                if n_reads > 10:  # In case EOF never hits because of bug.
+                    raise AssertionError(
+                        f"Failed to stop reading from file after {n_reads} loops. "
+                        f"Read {len(content_read)} bytes/characters. Expected {len(content)}."
+                    )
+            test_case.assertEqual(content_read, content)
+
+            # Go back to start of file and read the entire thing.
+            read_buffer.seek(0)
+            content_read = read_buffer.read()
+            test_case.assertEqual(content_read, content)
+
+            # Seek off the end of the file and should read empty back.
+            # We run this check twice since in some cases the handle will
+            # cache knowledge of the file size.
+            read_buffer.seek(1024)
+            test_case.assertEqual(read_buffer.tell(), 1024)
+            content_read = read_buffer.read()
+            test_case.assertEqual(len(content_read), 0, f"Read: {content_read!r}, expected empty.")
+
         # Write two copies of the content, overwriting the single copy there.
         with uri.open("w" + mode_suffix, **kwargs) as write_buffer:
             write_buffer.write(double_content)
@@ -116,32 +166,31 @@ def _check_open(
         uri.remove()
 
 
-class _GenericTestCase:
+if TYPE_CHECKING:
+
+    class TestCaseMixin(unittest.TestCase):
+        """Base class for mixin test classes that use TestCase methods."""
+
+        pass
+
+else:
+
+    class TestCaseMixin:
+        """Do-nothing definition of mixin base class for regular execution."""
+
+        pass
+
+
+class _GenericTestCase(TestCaseMixin):
     """Generic base class for test mixin."""
 
-    scheme: Optional[str] = None
-    netloc: Optional[str] = None
-    base_path: Optional[str] = None
+    scheme: str | None = None
+    netloc: str | None = None
+    base_path: str | None = None
     path1 = "test_dir"
     path2 = "file.txt"
 
-    # Because we use a mixin for tests mypy needs to understand that
-    # the unittest.TestCase methods exist.
-    # We do not inherit from unittest.TestCase because that results
-    # in the tests defined here being run as well as the tests in the
-    # test file itself. We can make those tests skip but it gives an
-    # uniformative view of how many tests are running.
-    assertEqual: Callable
-    assertNotEqual: Callable
-    assertIsNone: Callable
-    assertIn: Callable
-    assertNotIn: Callable
-    assertFalse: Callable
-    assertTrue: Callable
-    assertRaises: Callable
-    assertLogs: Callable
-
-    def _make_uri(self, path: str, netloc: Optional[str] = None) -> str:
+    def _make_uri(self, path: str, netloc: str | None = None) -> str:
         if self.scheme is not None:
             if netloc is None:
                 netloc = self.netloc
@@ -156,7 +205,7 @@ class _GenericTestCase:
 
 
 class GenericTestCase(_GenericTestCase):
-    """Test cases for generic manipulation of a `ResourcePath`"""
+    """Test cases for generic manipulation of a `ResourcePath`."""
 
     def setUp(self) -> None:
         if self.scheme is None:
@@ -208,7 +257,7 @@ class GenericTestCase(_GenericTestCase):
         self.assertEqual(extension.getExtension(), ".fits")
 
     def test_relative(self) -> None:
-        """Check that we can get subpaths back from two URIs"""
+        """Check that we can get subpaths back from two URIs."""
         parent = ResourcePath(self._make_uri(self.path1), forceDirectory=True)
         self.assertTrue(parent.isdir())
         child = parent.join("dir1/file.txt")
@@ -247,23 +296,25 @@ class GenericTestCase(_GenericTestCase):
 
         # Test with different netloc
         child = ResourcePath(self._make_uri("a/b/c.txt", netloc="my.host"))
-        parent = ResourcePath(self._make_uri("a", netloc="other"))
+        parent = ResourcePath(self._make_uri("a", netloc="other"), forceDirectory=True)
         self.assertIsNone(child.relative_to(parent), f"{child}.relative_to({parent})")
 
-        # Schemeless absolute child.
-        # Schemeless absolute URI is constructed using root= parameter.
-        # For now the root parameter can not be anything other than a file.
-        if self.scheme == "file":
-            parent = ResourcePath("/a/b/c", root=self.root_uri)
-            child = ResourcePath("d/e.txt", root=parent)
-            self.assertEqual(child.relative_to(parent), "d/e.txt", f"{child}.relative_to({parent})")
+        # This is an absolute path so will *always* return a file URI and
+        # ignore the root parameter.
+        parent = ResourcePath("/a/b/c", root=self.root_uri, forceDirectory=True)
+        self.assertEqual(parent.geturl(), "file:///a/b/c/")
 
-            parent = ResourcePath("c/", root="/a/b/")
-            self.assertEqual(child.relative_to(parent), "d/e.txt", f"{child}.relative_to({parent})")
+        parent = ResourcePath(self._make_uri("/a/b/c"), forceDirectory=True)
+        child = ResourcePath("d/e.txt", root=parent)
+        self.assertEqual(child.relative_to(parent), "d/e.txt", f"{child}.relative_to({parent})")
 
-            # Absolute schemeless child with relative parent will always fail.
-            parent = ResourcePath("d/e.txt", forceAbsolute=False)
-            self.assertIsNone(child.relative_to(parent), f"{child}.relative_to({parent})")
+        parent = ResourcePath("c/", root=ResourcePath(self._make_uri("/a/b/")))
+        self.assertEqual(child.relative_to(parent), "d/e.txt", f"{child}.relative_to({parent})")
+
+        # Absolute schemeless child with relative parent will always fail.
+        child = ResourcePath("d/e.txt", root="/a/b/c")
+        parent = ResourcePath("d/e.txt", forceAbsolute=False)
+        self.assertIsNone(child.relative_to(parent), f"{child}.relative_to({parent})")
 
     def test_parents(self) -> None:
         """Test of splitting and parent walking."""
@@ -282,7 +333,7 @@ class GenericTestCase(_GenericTestCase):
         self.assertEqual(child_file.parent().parent(), parent)
 
     def test_escapes(self) -> None:
-        """Special characters in file paths"""
+        """Special characters in file paths."""
         src = self.root_uri.join("bbb/???/test.txt")
         self.assertNotIn("???", src.path)
         self.assertIn("???", src.unquoted_path)
@@ -387,8 +438,9 @@ class GenericTestCase(_GenericTestCase):
         )
         self.assertEqual(joined, ResourcePath(f"{self.root}hsc/payload/test.qgraph"))
 
-        with self.assertRaises(ValueError):
-            ResourcePath(f"{self.root}hsc/payload/").join(ResourcePath("test.qgraph"))
+        qgraph = ResourcePath("test.qgraph")  # Absolute URI
+        joined = ResourcePath(f"{self.root}hsc/payload/").join(qgraph)
+        self.assertEqual(joined, qgraph)
 
     def test_quoting(self) -> None:
         """Check that quoting works."""
@@ -406,21 +458,21 @@ class GenericTestCase(_GenericTestCase):
         """Check that greater/less comparison operators work."""
         a = self._make_uri("a.txt")
         b = self._make_uri("b/")
-        self.assertTrue(a < b)
+        self.assertLess(a, b)
         self.assertFalse(a < a)
-        self.assertTrue(a <= b)
-        self.assertTrue(a <= a)
-        self.assertTrue(b > a)
+        self.assertLessEqual(a, b)
+        self.assertLessEqual(a, a)
+        self.assertGreater(b, a)
         self.assertFalse(b > b)
-        self.assertTrue(b >= a)
-        self.assertTrue(b >= b)
+        self.assertGreaterEqual(b, a)
+        self.assertGreaterEqual(b, b)
 
 
 class GenericReadWriteTestCase(_GenericTestCase):
     """Test schemes that can read and write using concrete resources."""
 
-    transfer_modes = ("copy", "move")
-    testdir: Optional[str] = None
+    transfer_modes: tuple[str, ...] = ("copy", "move")
+    testdir: str | None = None
 
     def setUp(self) -> None:
         if self.scheme is None:
@@ -441,9 +493,8 @@ class GenericReadWriteTestCase(_GenericTestCase):
             self.tmpdir.mkdir()
 
     def tearDown(self) -> None:
-        if self.tmpdir:
-            if self.tmpdir.isLocal:
-                removeTestTempDir(self.tmpdir.ospath)
+        if self.tmpdir and self.tmpdir.isLocal:
+            removeTestTempDir(self.tmpdir.ospath)
 
     def test_file(self) -> None:
         uri = self.tmpdir.join("test.txt")
@@ -487,7 +538,7 @@ class GenericReadWriteTestCase(_GenericTestCase):
         self.assertEqual(newdir.size(), 0)
 
         newfile = newdir.join("temp.txt")
-        newfile.write("Data".encode())
+        newfile.write(b"Data")
         self.assertTrue(newfile.exists())
 
         file = self.tmpdir.join("file.txt")
@@ -584,27 +635,28 @@ class GenericReadWriteTestCase(_GenericTestCase):
 
         # Temporary (possibly remote) resource.
         # Transfers between temporary resources.
-        with ResourcePath.temporary_uri(prefix=self.tmpdir.join("tmp"), suffix=".json") as remote_tmp:
-            # Temporary local resource.
-            with ResourcePath.temporary_uri(suffix=".json") as local_tmp:
-                remote_tmp.write(b"42")
-                if not remote_tmp.isLocal:
-                    for transfer in ("link", "symlink", "hardlink", "relsymlink"):
-                        with self.assertRaises(RuntimeError):
-                            # Trying to symlink a remote resource is not going
-                            # to work. A hardlink could work but would rely
-                            # on the local temp space being on the same
-                            # filesystem as the target.
-                            local_tmp.transfer_from(remote_tmp, transfer)
-                local_tmp.transfer_from(remote_tmp, "move")
-                self.assertFalse(remote_tmp.exists())
-                remote_tmp.transfer_from(local_tmp, "auto", overwrite=True)
-                self.assertEqual(local_tmp.read(), remote_tmp.read())
+        with (
+            ResourcePath.temporary_uri(prefix=self.tmpdir.join("tmp"), suffix=".json") as remote_tmp,
+            ResourcePath.temporary_uri(suffix=".json") as local_tmp,
+        ):
+            remote_tmp.write(b"42")
+            if not remote_tmp.isLocal:
+                for transfer in ("link", "symlink", "hardlink", "relsymlink"):
+                    with self.assertRaises(RuntimeError):
+                        # Trying to symlink a remote resource is not going
+                        # to work. A hardlink could work but would rely
+                        # on the local temp space being on the same
+                        # filesystem as the target.
+                        local_tmp.transfer_from(remote_tmp, transfer)
+            local_tmp.transfer_from(remote_tmp, "move")
+            self.assertFalse(remote_tmp.exists())
+            remote_tmp.transfer_from(local_tmp, "auto", overwrite=True)
+            self.assertEqual(local_tmp.read(), remote_tmp.read())
 
-                # Transfer of missing remote.
-                remote_tmp.remove()
-                with self.assertRaises(FileNotFoundError):
-                    local_tmp.transfer_from(remote_tmp, "auto", overwrite=True)
+            # Transfer of missing remote.
+            remote_tmp.remove()
+            with self.assertRaises(FileNotFoundError):
+                local_tmp.transfer_from(remote_tmp, "auto", overwrite=True)
 
     def test_local(self) -> None:
         """Check that remote resources can be made local."""
@@ -659,7 +711,7 @@ class GenericReadWriteTestCase(_GenericTestCase):
         self.assertEqual(found, expected_yaml)
 
         # Now two explicit directories and a file
-        expected = set(u for u in expected_yaml)
+        expected = set(expected_yaml)
         expected.add(file)
 
         found = set(
@@ -691,7 +743,7 @@ class GenericReadWriteTestCase(_GenericTestCase):
         # at the end
         file2 = root.join("config/templates/templates-bad.yaml")
         found_grouped = [
-            [uri for uri in group]
+            list(group)
             for group in ResourcePath.findFileResources([file, file2, root.join("dir2/")], grouped=True)
             if not isinstance(group, ResourcePath)  # For mypy.
         ]
@@ -736,11 +788,11 @@ class GenericReadWriteTestCase(_GenericTestCase):
         # Again with grouping.
         # (mypy gets upset not knowing which of the two options is being
         # returned so add useless instance check).
-        found_list = list(
-            [uri for uri in group]
+        found_list = [
+            list(group)
             for group in ResourcePath.findFileResources([root], grouped=True)
             if not isinstance(group, ResourcePath)  # For mypy.
-        )
+        ]
         self.assertEqual(len(found_list), 2)
         self.assertEqual(len(found_list[0]), n_dir1)
         self.assertEqual(len(found_list[1]), n_dir2)
