@@ -22,8 +22,9 @@ import tempfile
 import time
 import unittest
 import warnings
+from collections.abc import Callable
 from threading import Thread
-from typing import Callable, Tuple, cast
+from typing import cast
 
 try:
     from cheroot import wsgi
@@ -50,6 +51,8 @@ TESTDIR = os.path.abspath(os.path.dirname(__file__))
 
 
 class GenericHttpTestCase(GenericTestCase, unittest.TestCase):
+    """Generic tests of http URIs."""
+
     scheme = "http"
     netloc = "server.example"
 
@@ -171,6 +174,9 @@ class HttpReadWriteWebdavTestCase(GenericReadWriteTestCase, unittest.TestCase):
             self.assertIsNotNone(handle._completeBuffer)
             self.assertEqual(result, contents)
 
+            # Check that flush works on read-only handle.
+            handle.flush()
+
         # Verify reading as a string handle works as expected.
         with remote_file.open("r") as handle:
             self.assertIsInstance(handle, io.TextIOWrapper)
@@ -182,6 +188,9 @@ class HttpReadWriteWebdavTestCase(GenericReadWriteTestCase, unittest.TestCase):
             result = handle.read()
             self.assertEqual(result, contents)
 
+            # Check that flush works on read-only handle.
+            handle.flush()
+
         # Verify that write modes invoke the default base method
         with remote_file.open("w") as handle:
             self.assertIsInstance(handle, io.StringIO)
@@ -457,7 +466,7 @@ class HttpReadWriteWebdavTestCase(GenericReadWriteTestCase, unittest.TestCase):
         """Return a randomly selected name for a file"""
         return cls._get_name(prefix="file")
 
-    def _generate_file(self, remove_when_done=True) -> Tuple[str, int]:
+    def _generate_file(self, remove_when_done=True) -> tuple[str, int]:
         """Create a local file of random size with random contents.
 
         Returns
@@ -504,7 +513,7 @@ class HttpResourcePathConfigTestCase(unittest.TestCase):
 
     def test_send_expect_header(self):
         # Ensure environment variable LSST_HTTP_PUT_SEND_EXPECT_HEADER is
-        # inspected to initialize the HttpResourcePath config class.
+        # inspected to initialize the HttpResourcePathConfig config class.
         with unittest.mock.patch.dict(os.environ, {}, clear=True):
             importlib.reload(lsst.resources.http)
             config = HttpResourcePathConfig()
@@ -515,20 +524,31 @@ class HttpResourcePathConfigTestCase(unittest.TestCase):
             config = HttpResourcePathConfig()
             self.assertTrue(config.send_expect_on_put)
 
+    def test_collect_memory_usage(self):
+        # Ensure environment variable LSST_HTTP_COLLECT_MEMORY_USAGE is
+        # inspected to initialize the HttpResourcePathConfig class.
+        with unittest.mock.patch.dict(os.environ, {}, clear=True):
+            importlib.reload(lsst.resources.http)
+            config = HttpResourcePathConfig()
+            self.assertFalse(config.collect_memory_usage)
+
+        with unittest.mock.patch.dict(os.environ, {"LSST_HTTP_COLLECT_MEMORY_USAGE": "true"}, clear=True):
+            importlib.reload(lsst.resources.http)
+            config = HttpResourcePathConfig()
+            self.assertTrue(config.collect_memory_usage)
+
     def test_timeout(self):
         # Ensure that when the connect and read timeouts are not specified
         # the default values are stored in the config.
         with unittest.mock.patch.dict(os.environ, {}, clear=True):
             importlib.reload(lsst.resources.http)
             config = HttpResourcePathConfig()
-            self.assertEqual(
-                config.timeout,
-                (lsst.resources.http.DEFAULT_TIMEOUT_CONNECT, lsst.resources.http.DEFAULT_TIMEOUT_READ),
-            )
+            self.assertAlmostEqual(config.timeout[0], config.DEFAULT_TIMEOUT_CONNECT)
+            self.assertAlmostEqual(config.timeout[1], config.DEFAULT_TIMEOUT_READ)
 
         # Ensure that when both the connect and read timeouts are specified
-        # they are stored in the config.
-        connect_timeout, read_timeout = 100, 100
+        # they are both stored in the config.
+        connect_timeout, read_timeout = 100.5, 200.8
         with unittest.mock.patch.dict(
             os.environ,
             {"LSST_HTTP_TIMEOUT_CONNECT": str(connect_timeout), "LSST_HTTP_TIMEOUT_READ": str(read_timeout)},
@@ -537,7 +557,22 @@ class HttpResourcePathConfigTestCase(unittest.TestCase):
             # Force module reload.
             importlib.reload(lsst.resources.http)
             config = HttpResourcePathConfig()
-            self.assertEqual(config.timeout, (connect_timeout, read_timeout))
+            self.assertAlmostEqual(config.timeout[0], connect_timeout)
+            self.assertAlmostEqual(config.timeout[1], read_timeout)
+
+        # Ensure that invalid float values (including NaN values) raise a
+        # ValueError.
+        for value in ("invalid", "NaN"):
+            with unittest.mock.patch.dict(
+                os.environ,
+                {"LSST_HTTP_TIMEOUT_CONNECT": value, "LSST_HTTP_TIMEOUT_READ": value},
+                clear=True,
+            ):
+                # Force module reload.
+                importlib.reload(lsst.resources.http)
+                with self.assertRaises(ValueError):
+                    config = HttpResourcePathConfig()
+                    config.timeout()
 
     def test_front_end_connections(self):
         # Ensure that when the number of front end connections is not specified
@@ -545,9 +580,7 @@ class HttpResourcePathConfigTestCase(unittest.TestCase):
         with unittest.mock.patch.dict(os.environ, {}, clear=True):
             importlib.reload(lsst.resources.http)
             config = HttpResourcePathConfig()
-            self.assertEqual(
-                config.front_end_connections, int(lsst.resources.http.DEFAULT_FRONTEND_PERSISTENT_CONNECTIONS)
-            )
+            self.assertEqual(config.front_end_connections, config.DEFAULT_FRONTEND_PERSISTENT_CONNECTIONS)
 
         # Ensure that when the number of front end connections is specified
         # it is stored in the config.
@@ -565,9 +598,7 @@ class HttpResourcePathConfigTestCase(unittest.TestCase):
         with unittest.mock.patch.dict(os.environ, {}, clear=True):
             importlib.reload(lsst.resources.http)
             config = HttpResourcePathConfig()
-            self.assertEqual(
-                config.back_end_connections, int(lsst.resources.http.DEFAULT_BACKEND_PERSISTENT_CONNECTIONS)
-            )
+            self.assertEqual(config.back_end_connections, config.DEFAULT_BACKEND_PERSISTENT_CONNECTIONS)
 
         # Ensure that when the number of back end connections is specified
         # it is stored in the config.
@@ -595,12 +626,52 @@ class HttpResourcePathConfigTestCase(unittest.TestCase):
             self.assertEqual(config.digest_algorithm, "")
 
         # Ensure that an accepted digest algorithm is stored.
-        for digest in lsst.resources.http.ACCEPTED_DIGESTS:
+        for digest in HttpResourcePathConfig().ACCEPTED_DIGESTS:
             with unittest.mock.patch.dict(os.environ, {"LSST_HTTP_DIGEST": digest}, clear=True):
                 importlib.reload(lsst.resources.http)
                 config = HttpResourcePathConfig()
                 self.assertTrue(config.digest_algorithm, digest)
 
+    def test_backoff_interval(self):
+        # Ensure that when no backoff interval is defined, the default values
+        # are used.
+        with unittest.mock.patch.dict(os.environ, {}, clear=True):
+            importlib.reload(lsst.resources.http)
+            config = HttpResourcePathConfig()
+            self.assertAlmostEqual(config.backoff_min, config.DEFAULT_BACKOFF_MIN)
+            self.assertAlmostEqual(config.backoff_max, config.DEFAULT_BACKOFF_MAX)
+
+        # Ensure that an invalid value for backoff interval is ignored and
+        # the default value is used.
+        with unittest.mock.patch.dict(
+            os.environ, {"LSST_HTTP_BACKOFF_MIN": "XXX", "LSST_HTTP_BACKOFF_MAX": "YYY"}, clear=True
+        ):
+            importlib.reload(lsst.resources.http)
+            config = HttpResourcePathConfig()
+            self.assertAlmostEqual(config.backoff_min, config.DEFAULT_BACKOFF_MIN)
+            self.assertAlmostEqual(config.backoff_max, config.DEFAULT_BACKOFF_MAX)
+
+        # Ensure that NaN values are ignored and the defaults values are used.
+        with unittest.mock.patch.dict(
+            os.environ, {"LSST_HTTP_BACKOFF_MIN": "NaN", "LSST_HTTP_BACKOFF_MAX": "NaN"}, clear=True
+        ):
+            importlib.reload(lsst.resources.http)
+            config = HttpResourcePathConfig()
+            self.assertAlmostEqual(config.backoff_min, config.DEFAULT_BACKOFF_MIN)
+            self.assertAlmostEqual(config.backoff_max, config.DEFAULT_BACKOFF_MAX)
+
+        # Ensure that when specified, valid limits backoff interval are used.
+        backoff_min, backoff_max = 3.0, 8.0
+        with unittest.mock.patch.dict(
+            os.environ,
+            {"LSST_HTTP_BACKOFF_MIN": str(backoff_min), "LSST_HTTP_BACKOFF_MAX": str(backoff_max)},
+            clear=True,
+        ):
+            importlib.reload(lsst.resources.http)
+            config = HttpResourcePathConfig()
+            self.assertAlmostEqual(config.backoff_min, backoff_min)
+            self.assertAlmostEqual(config.backoff_max, backoff_max)
+
 
 class WebdavUtilsTestCase(unittest.TestCase):
     """Test for the Webdav related utilities."""
@@ -609,9 +680,8 @@ class WebdavUtilsTestCase(unittest.TestCase):
         self.tmpdir = ResourcePath(makeTestTempDir(TESTDIR))
 
     def tearDown(self):
-        if self.tmpdir:
-            if self.tmpdir.isLocal:
-                removeTestTempDir(self.tmpdir.ospath)
+        if self.tmpdir and self.tmpdir.isLocal:
+            removeTestTempDir(self.tmpdir.ospath)
 
     @responses.activate
     def test_is_webdav_endpoint(self):
@@ -623,6 +693,10 @@ class WebdavUtilsTestCase(unittest.TestCase):
         responses.add(responses.OPTIONS, plainHttpEndpoint, status=200)
         self.assertFalse(_is_webdav_endpoint(plainHttpEndpoint))
 
+        notWebdavEndpoint = "http://www.notwebdav.org"
+        responses.add(responses.OPTIONS, notWebdavEndpoint, status=403)
+        self.assertFalse(_is_webdav_endpoint(notWebdavEndpoint))
+
     def test_is_protected(self):
         self.assertFalse(_is_protected("/this-file-does-not-exist"))
 
@@ -668,6 +742,15 @@ class BearerTokenAuthTestCase(unittest.TestCase):
         req = auth(requests.Request("GET", "https://example.org").prepare())
         self.assertEqual(req.headers.get("Authorization"), f"Bearer {self.token}")
 
+    def test_token_insecure_http(self):
+        """Ensure that no 'Authorization' header is attached to a request when
+        using insecure HTTP.
+        """
+        auth = BearerTokenAuth(self.token)
+        for url in ("http://example.org", "HTTP://example.org", "HttP://example.org"):
+            req = auth(requests.Request("GET", url).prepare())
+            self.assertIsNone(req.headers.get("Authorization"))
+
     def test_token_file(self):
         """Ensure when the provided token is a file path, its contents is
         correctly used in the the 'Authorization' header of the requests.
@@ -718,7 +801,6 @@ class SessionStoreTestCase(unittest.TestCase):
         """Ensure if user certificate and private key are provided, they are
         used for authenticating the client.
         """
-
         # Create mock certificate and private key files.
         with tempfile.NamedTemporaryFile(mode="wt", dir=self.tmpdir.ospath, delete=False) as f:
             f.write("CERT")
@@ -771,7 +853,6 @@ class SessionStoreTestCase(unittest.TestCase):
 
     def test_sessions(self):
         """Ensure the session caching mechanism works."""
-
         # Ensure the store provides a session for a given URL
         root_url = "https://example.org"
         store = SessionStore()
