@@ -19,34 +19,23 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+import os
 import unittest
-
-from lsst.resources.s3utils import clean_test_environment
+from unittest import mock
 
 try:
     import boto3
+    from botocore.exceptions import ParamValidationError
     from moto import mock_s3
 except ImportError:
     boto3 = None
 
-    def mock_s3(cls):
-        """A no-op decorator in case moto mock_s3 can not be imported."""
-        return cls
-
-
 from lsst.resources import ResourcePath
 from lsst.resources.location import Location
-from lsst.resources.s3utils import (
-    bucketExists,
-    getS3Client,
-    s3CheckFileExists,
-    setAwsEnvCredentials,
-    unsetAwsEnvCredentials,
-)
+from lsst.resources.s3utils import bucketExists, clean_test_environment_for_s3, getS3Client, s3CheckFileExists
 
 
 @unittest.skipIf(not boto3, "Warning: boto3 AWS SDK not found!")
-@mock_s3
 class S3UtilsTestCase(unittest.TestCase):
     """Test for the S3 related utilities."""
 
@@ -54,10 +43,8 @@ class S3UtilsTestCase(unittest.TestCase):
     fileName = "testFileName"
 
     def setUp(self):
-        # set up some fake credentials if they do not exist
-        self.usingDummyCredentials = setAwsEnvCredentials()
-
-        clean_test_environment(self)
+        self.enterContext(clean_test_environment_for_s3())
+        self.enterContext(mock_s3())
 
         self.client = getS3Client()
         try:
@@ -74,14 +61,21 @@ class S3UtilsTestCase(unittest.TestCase):
 
         self.client.delete_bucket(Bucket=self.bucketName)
 
-        # unset any potentially set dummy credentials
-        if self.usingDummyCredentials:
-            unsetAwsEnvCredentials()
-
     def testBucketExists(self):
         self.assertTrue(bucketExists(f"{self.bucketName}"))
         self.assertFalse(bucketExists(f"{self.bucketName}_no_exist"))
 
+    def testCephBucket(self):
+        with mock.patch.dict(os.environ, {"LSST_DISABLE_BUCKET_VALIDATION": "N"}):
+            self.assertEqual(os.environ["LSST_DISABLE_BUCKET_VALIDATION"], "N")
+            local_client = getS3Client()
+            with self.assertRaises(ParamValidationError):
+                bucketExists("foo:bar", local_client)
+        with mock.patch.dict(os.environ, {"LSST_DISABLE_BUCKET_VALIDATION": "1"}):
+            self.assertEqual(os.environ["LSST_DISABLE_BUCKET_VALIDATION"], "1")
+            local_client = getS3Client()
+            self.assertFalse(bucketExists("foo:bar", local_client))
+
     def testFileExists(self):
         self.assertTrue(s3CheckFileExists(client=self.client, bucket=self.bucketName, path=self.fileName)[0])
         self.assertFalse(
